#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version
#
#
# HPC Anaconda/Miniforge module wrapper
#
# The purpose of this wrapper is to allow a functional packaging of Anaconda/Miniforge in a module
# that when loaded allows 'conda <commands...>' and 'mamba <commands...>' without any need for
# 'conda init' etc.
#
# It also:
#
# - Avoids autoactivation of an environment at load (e.g., so users gets a conda environment loaded for every shell.
# - Blocks 'conda init' to avoid users ending up with conda commands in their init files that does not work.
# - Blocks attempts to modify the base environment.
# - Optionally blocks attempts to activate the base environemnt (set HPC_CONDA_BLOCK_BASE to enable)


BINNAME="${FUNCNAME[1]}"

if [ -z "${HPC_CONDA_PATH-}" -o -z "${HPC_CONDA_WRAPPATH-}" ]; then
    echo "${BINNAME}: command not found" >&2
    return
else
    CONDA_PATH="$HPC_CONDA_PATH"
fi

if [ "$BINNAME" != "conda" -a "$BINNAME" != "mamba" ]; then
    echo "${BINNAME}: command not found" >&2
    return
fi

if [ "${1-}" == "env" ]; then
    COMMAND="${2-}"
else
    COMMAND="${1-}"
fi

# If we are in an active environment that isn't base, just run the requested conda command
if [ -z "$COMMAND" -o -n "${CONDA_DEFAULT_ENV-}" -a "${CONDA_DEFAULT_ENV-}" != "base" ]; then
    if type "hpc_orig_${BINNAME}" >& /dev/null; then
    
	"hpc_orig_${BINNAME}" "$@"

	# if the command exits the conda environement, unload everything
	if [ -z "${CONDA_DEFAULT_ENV-}" ]; then
	    "hpc_orig_${BINNAME}" deactivate
	    unset -f hpc_orig_conda
	    unset -f hpc_orig_mamba
	    # Clean any leftovers in the user's PATH (usually condabin)
	    export PATH="$(awk -F: -vOFS=':' '{for(i=1;i<=NF;i++) {if ($i !~ "^'"$HPC_CONDA_PATH"'") {j+=1; $j=$i}} NF=j; print $0}'<<<"$PATH")"
	fi
	return
    fi
fi

# If we are in the base environment, allow specifically only to deactivate it
if [ "${CONDA_DEFAULT_ENV-}" == "base" -a "$1" == "deactivate" ]; then
    "hpc_orig_${BINNAME}" deactivate
    unset -f hpc_orig_conda
    unset -f hpc_orig_mamba
    # Clean any leftovers in the user's PATH (usually condabin)
    export PATH="$(awk -F: -vOFS=':' '{for(i=1;i<=NF;i++) {if ($i !~ "^'"$HPC_CONDA_PATH"'") {j+=1; $j=$i}} NF=j; print $0}'<<<"$PATH")"
    return
fi

function help_message() {
    echo "Our setup of the ${BINNAME} command REQUIRES an active environment before"
    echo "commands such as '${BINNAME} install', '${BINNAME} update', etc."
    echo "(You cannot customize the 'base' environment.)"
    echo ""
    echo "For example, to set up a customized environment with python 3.8 and a recent version of scipy:"
    echo ""
    echo "  ${BINNAME} create -n myenv python=3.8 scipy"
    echo "  ${BINNAME} activate myenv"
    echo ""
    echo "If you get this message when trying to follow an external help document"
    echo "that instructs you to run '${BINNAME} install' etc., just first create a new environment"
    echo "and activate it using the instructions above, and then continue onwards with"
    echo "the instructions."
    echo ""
    echo "For more details about managing conda environments, see the following link:"
    echo ""
    echo "  https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-environments.html"
    echo ""
}

# Without an active environment, block install commands without a specific environment to install to
if [ "$COMMAND" == "install" -o "$COMMAND" == "update" ]; then
    ENVARG="FALSE"
    for ARG in "$@"; do
	if [ "$ARG" == "-n" -o "$ARG" == "--name" -o "$ARG" == "--prefix" -o "$ARG" == "-p" ]; then
	    ENVARG="TRUE"
	fi   
    done
    if [ "$ENVARG" != "TRUE" ]; then
	help_message
	return
    fi
fi

if [ "${1-}" == "init" ]; then
    echo "In our setup of the ${BINNAME} command you do not need to run '${BINNAME} init'. Just:"
    echo ""
    echo "  1. First use 'module load' to load a module providing ${BINNAME} (you already did this)"
    echo "  2. Then activate an environment, e.g., '${BINNAME} activate myenv' (where you must have created myenv using '${BINNAME} create -n myenv' first)"
    echo ""
    return
fi

if [ "${1-}" == "activate" -a -z "${2-}" -a -n "${HPC_CONDA_BLOCK_BASE-}" ]; then 
    echo "You need to specify a custom environment to activate, e.g.,:"
    echo ""
    echo "  ${BINNAME} activate myenv"
    echo ""
    echo "(where you must have created myenv using '${BINNAME} create -n myenv' first)"
    echo ""
    return
fi

if [ "${1-}" == "activate" ]; then

    export CONDA_AUTO_ACTIVATE_BASE=false
    if [ -e "${CONDA_PATH}/bin/conda" ]; then
	unset -f conda >& /dev/null
	eval "$("${CONDA_PATH}/bin/conda" shell.bash hook)"
	eval hpc_orig_"$(declare -f conda)"
	function conda() { source "$HPC_CONDA_WRAPPATH/hpc_conda_wrap.sh" "$@"; }
    fi

    if [ -e "${CONDA_PATH}/etc/profile.d/mamba.sh" ]; then    
	unset -f mamba >& /dev/null
	source "${CONDA_PATH}/etc/profile.d/mamba.sh"
	eval hpc_orig_"$(declare -f mamba)"
	function mamba() { source "$HPC_CONDA_WRAPPATH/hpc_conda_wrap.sh" "$@"; }
    fi
	
    "hpc_orig_${BINNAME}" "$@"

    # if the environment failed to activate, unload everything again
    if [ -z "${CONDA_DEFAULT_ENV-}" ]; then
	"hpc_orig_${BINNAME}" deactivate
	unset -f hpc_orig_conda
	unset -f hpc_orig_mamba
	# Clean any leftovers in the user's PATH (usually condabin)
	export PATH="$(awk -F: -vOFS=':' '{for(i=1;i<=NF;i++) {if ($i !~ "^'"$HPC_CONDA_PATH"'") {j+=1; $j=$i}} NF=j; print $0}'<<<"$PATH")"
    fi
    return
else
    # Run the command the user wants, but protected by a subshell so it cannot modify the environment
    (
	export CONDA_AUTO_ACTIVATE_BASE=false
	unset -f conda >& /dev/null
	unset -f mamba >& /dev/null

	if [ -e "${CONDA_PATH}/bin/conda" ]; then
	    eval "$("${CONDA_PATH}/bin/conda" shell.bash hook)"
	fi

	if [ -e "${CONDA_PATH}/etc/profile.d/mamba.sh" ]; then    
	    source "${CONDA_PATH}/etc/profile.d/mamba.sh"
	fi

	"$BINNAME" "$@"
    )
    return
fi
