#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version
#
#
# HPC pip wrapper for Anaconda/Miniforge modules
#
# The purpose of this wrapper is to only allow users to run pip inside a custom environment once an
# Anaconda/Miniforge module is loaded.
# (i.e., not without an environment activated, nor in the base environment)


BINNAME="${FUNCNAME[1]}"

if [ -z "${HPC_CONDA_PATH-}" ]; then
    "$BINNAME" "$@"
    return
fi

if [ -n "${CONDA_DEFAULT_ENV-}" -a "${CONDA_DEFAULT_ENV-}" != "base" -a -n "${CONDA_PREFIX-}" ]; then

    # If we are in an active conda environment where the pip command (pip/pip2/pip3) the user asks for, just run it
    if [ -e "${CONDA_PREFIX-}/bin/$BINNAME" ]; then
	"${CONDA_PREFIX-}/bin/$BINNAME" "$@"
	return
    fi

    # Educate the user about prefering conda channels
    echo "" >&2
    echo "Note: you are trying to run $BINNAME in a conda environment that does not have that command." >&2
    echo "Before installing any software with $BINNAME, please check if the same software is available via one" >&2
    echo "of the existing anaconda channels:" >&2
    echo "- by running, e.g.,: conda search -c conda-forge name_of_package" >&2
    echo "- or by using the online conda package search here: https://anaconda.org/search" >&2
    echo "" >&2
    if [ -e "${CONDA_PREFIX-}/bin/pip" ]; then
	# If the active conda environment has 'pip', point them to that 
	echo "Your conda environment supplies the command 'pip' without a version number." >&2
	echo "Hence, if you need to use pip to install the software, use that command." >&2
    else
	# If the active conda environment does not have 'pip', help them get an environment with pip
	echo "If you still need to install software with pip," >&2
	echo "you are recommended to either add pip inside this environment or (recommended) re-create a new environment with the same packages + pip." >&2
	echo "" >&2
	echo "You can try to copy the present environment and add pip as follows" >&2
	echo "(substitute 'newenv' with the new name to use):" >&2
	echo "-------------------------------------------------------------" >&2
	echo "              conda list -e > env_reqs.txt" >&2
	echo "              conda create -n newenv --file env_reqs.txt pip" >&2
	echo "              conda activate newenv" >&2
	echo "-------------------------------------------------------------" >&2
	echo ")" >&2
	echo "(We generally do NOT RECOMMENDED adding packages to existing conda environements, since it may" >&2
	echo "alter them unexpectedly, e.g., by downgrading library versions or even the version of python" >&2
        echo "itself. Nevertheless, if you are confident that you can handle this, you can instead alternatively" >&2
	echo "try to add pip to the present environment by simply:" >&2
	echo "  conda install pip" >&2
	echo ")" >&2
    fi
    echo >&2
    echo "$BINNAME: command not found" >&2
    return
fi

echo "Note: you are trying to run $BINNAME while having a module that provides conda loaded, but while not having" >&2
echo "an activate conda environment. Please first activate the environment you want to use, e.g., by:" >&2
echo >&2
echo "  conda activate envname" >&2
echo >&2
echo "(You should generally NEVER use pip outside of your conda environments as that can lead to very inconsistent python dependency situations."
echo "However, if you are absolutely sure this is what you want to do, you can call the pip binary you want to use,"
echo "e.g., the system pip, using its full absolute path.)"
echo >&2
echo "$BINNAME: command not found" >&2
return
