#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version
#
#  
# This script installs an Anaconda module into the ssetools curated module directory
#


set -e

module load buildtool-easybuild

VERSION="2023.09-0"

EASYBUILD_PREFIX="/software/sse2/${HPC_SSE_SYSNAME}/easybuild/pure" eb "IncludedFromEasyBuild/Anaconda3-${VERSION}.eb" --robot

if [ ! -e "/software/sse2/${HPC_SSE_SYSNAME}/modules" ]; then
   echo "Missing the module directory in which to install Anaconda?"
   exit 1
fi 

mkdir -p "/software/sse2/${HPC_SSE_SYSNAME}/modules/Anaconda"
cat > "/software/sse2/${HPC_SSE_SYSNAME}/modules/Anaconda/.${VERSION}-hpc1.lua" <<EOF
help([==[

Description
===========
Built to complement the rich, open source Python community,
the Anaconda platform provides an enterprise-ready data analytics platform 
that empowers companies to adopt a modern open data science analytics architecture.

Loading this module only makes available the 'conda' command, with which 
anaconda environments can be created and activated.


More information
================
 - Homepage: https://www.continuum.io/anaconda-overview
]==])

whatis([==[Description: Built to complement the rich, open source Python community,
the Anaconda platform provides an enterprise-ready data analytics platform 
that empowers companies to adopt a modern open data science analytics architecture.
]==])
whatis([==[Homepage: https://www.continuum.io/anaconda-overview]==])

local root = "/software/sse2/" .. os.getenv("HPC_SSE_SYSNAME") .. "/easybuild/pure/software/Anaconda3/${VERSION}"
local wrappath = "/software/sse2/generic/manual/ssetools/" .. os.getenv("HPC_SSE_VER") .. "/conda"

setenv("HPC_CONDA_PATH",root)
setenv("HPC_CONDA_WRAPPATH",wrappath)

set_shell_function("conda", "source " .. wrappath .. "/hpc_conda_wrap.sh")
set_shell_function("pip", "source " .. wrappath .. "/hpc_pip_wrap.sh")

conflict("Python")

nsc_message_on_load(
[==[
**********************************
You have loaded an Anaconda module
**********************************

JUST LOADING THIS MODULE DOES *NOT* CHANGE THE PYTHON VERSION, NOR ENABLES ANY EXTRA PYTHON MODULES.

This module makes available the 'conda' command, with which users can create
named "conda environments" and then activate them. 

To create and use a customized environment use 'conda create ...' and 
then 'conda activate ...' (see example below).

---------------------------------------------------------------------------------------------------
NOTE: We strongly advice against placing 'conda activate' (with or without arguments) 
      in your shell initialization files, (e.g. '.bashrc' or '.bash_profile') since this 
      severly alters the environment for running software in ways that cause unpredictable 
      issues that can be difficult to diagnose.
---------------------------------------------------------------------------------------------------

Example usage:

  * Setting up a customized conda environment and run a Python program in it:

    conda create -n myenv python=3.8 scipy=1.5.2
    conda activate myenv
    python my_scipy_python_program.py
 
  * To run the python program in the same environment when logging in the next time:

    conda activate myenv
    python my_scipy_python_program.py

More details on conda environment management here:

 https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-environments.html


]==])
EOF
