#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version
#
#  
# This script installs a Miniforge module into the ssetools curated module directory
#


set -e

module load buildtool-easybuild

VERSION="24.7.1-2"

EASYBUILD_PREFIX="/software/sse2/${HPC_SSE_SYSNAME}/easybuild/pure eb" "IncludedFromEasyBuild/Miniforge3-${VERSION}.eb" --robot --include-easyblocks=IncludedFromEasyBuild/miniforge.py

# Fix a configuration setting to avoid warnings about package downloading
cat >> "/software/sse2/${HPC_SSE_SYSNAME}/easybuild/pure/software/Miniforge3/${VERSION}/.condarc" <<EOF
pkgs_dirs:
- \$HOME/.conda/pkgs
EOF

if [ ! -e "/software/sse2/${HPC_SSE_SYSNAME}/modules" ]; then
   echo "Missing the module directory in which to install Miniforge?"
   exit 1
fi 

mkdir -p "/software/sse2/${HPC_SSE_SYSNAME}/modules/Miniforge"
cat > "/software/sse2/${HPC_SSE_SYSNAME}/modules/Miniforge/.${VERSION}-hpc1.lua" <<EOF
help([==[

Description
===========
Miniforge provides the "conda" and "mamba" tools to set up and manage conda runtime 
environments. The configuration of conda provided by this module uses the conda-forge 
channel for package installations. 

A conda runtime environment is a set of packages that provide precise versions of 
some desired software and its dependencies. These environments helps ensure reproducible 
behavior across different systems.

The Miniforge module provides both "conda" and "mamba". The latter is a faster, 
generally improved, drop-in replacement for "conda". You should be able to issue
the same commands with either "conda" or "mamba" interchangably.

(The difference between the this module, named "Miniforge", and setups using "miniconda"
distributed by Anaconda, is that while the latter provides essentially the same command
"conda", it is configured for using channels of packages provided by the Anaconda company, 
for which their terms-of-service apply which imposes some relevant restrictions.)

IMPORTANT: LOADING THIS MODULE DOES NOT BY ITSELF CHANGE YOUR PYTHON VERSION OR PROVIDE
ANY PYTHON PACKAGES. You must use the 'conda' or 'mamba' commands to create a conda 
environments and activate it to get access to any packaged software.

More information
================
 - Homepage: https://github.com/conda-forge/miniforge
]==])

whatis([==[Description: Miniforge provides the "conda" and "mamba" tools to set up and manage conda runtime environments.
]==])
whatis([==[Homepage: https://github.com/conda-forge/miniforge]==])

local root = "/software/sse2/" .. os.getenv("HPC_SSE_SYSNAME") .. "/easybuild/pure/software/Miniforge3/${VERSION}"
local wrappath = "/software/sse2/generic/manual/ssetools/" .. os.getenv("HPC_SSE_VER") .. "/conda"

setenv("HPC_CONDA_PATH",root)
setenv("HPC_CONDA_WRAPPATH",wrappath)

set_shell_function("conda", "source " .. wrappath .. "/hpc_conda_wrap.sh")
set_shell_function("mamba", "source " .. wrappath .. "/hpc_conda_wrap.sh")
set_shell_function("pip", "source " .. wrappath .. "/hpc_pip_wrap.sh")

conflict("Python")

nsc_message_on_load(
[==[
**********************************
You have loaded a Miniforge module
**********************************

JUST LOADING THIS MODULE DOES *NOT* CHANGE THE PYTHON VERSION, NOR ENABLES ANY EXTRA PYTHON MODULES.

This module makes available the 'mamba' and 'conda' commands from the miniforge distribution, 
with which users can create "conda environments" and then activate them. 

To create and use a customized environment use 'mamba create ...' and 
then 'mamba activate ...' (see example below).

---------------------------------------------------------------------------------------------------
NOTE: NSC strongly advices against placing 'mamba activate' (or 'conda activate', with or without 
      arguments)  in your shell initialization files, (e.g. '.bashrc' or '.bash_profile') since 
      this severly alters the environment for running software in ways that cause unpredictable 
      issues that can be difficult to diagnose.
---------------------------------------------------------------------------------------------------

Example usage:

  * Setting up a customized environment and run a Python program in it:

    mamba create -n myenv python=3.8 scipy=1.5.2
    mamba activate myenv
    python my_scipy_python_program.py
 
  * To run the python program in the same environment when logging in the next time:

    mabma activate myenv
    python my_scipy_python_program.py

More details on Anaconda environment management are available here:

  https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-environments.html

]==])
EOF
