#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version
#
# Script installing a new version of ssetools

VER=$(git describe --always --dirty=-devel)
VER_BARE=${VER#v}
LAST_GIT_TAG=$(git describe --tags --abbrev=0)

echo "Git describes the current version of the ssetools as: ${VER_BARE} (named: ${VER})"
echo "This should normally be a clean version number on the format: 1.2.3, if it is not, do not continue unless you know why."
echo ""
echo "Is the version described correctly? [y/n]"
read YN

if [ "$YN" != "y" ]; then
    echo ""
    echo "* To remove a -devel suffix make sure to have no edited uncommited files"
    echo "* To have a nice version number instread of a commit hash, you need to back-track to the most recent versioned release or create a new release."
    echo ""
    echo "To back-track to the most recent versioned release:"
    echo ""
    echo "    git checkout tags/${LAST_GIT_TAG}"
    echo ""
    echo "(Then run ./install.sh, and finally return to the head of the main branch by: git checkout main)"
    echo ""
    echo "Or, to instead create a new release based on the currect state of the repo:"
    echo ""
    echo "    git tag -s v<version number> -m v<version number>"
    echo ""
    echo "  Remember to use a v-prefix argument with a major, minor and patch number, e.g., v1.2.3)"
    echo ""
    exit 1
fi

if [ -e "/software/sse2/generic/manual/ssetools/${VER}" ]; then
    echo ""
    echo "The destination directory for this version already exist. If you mean to install over an older version, please remove: /software/sse2/generic/manual/ssetools/$VER"
    echo ""
    exit 1
fi

if [ -e "/software/sse2/generic/modules/hpc/.${VER_BARE}.lua" ]; then
    echo ""
    echo "The destination module for this version already exist."
    echo "If you mean to install over an older version, please remove this module first: /software/sse2/generic/modules/hpc/.${VER_BARE}.lua"
    echo ""
    exit 1
fi

rsync -a ./ "/software/sse2/generic/manual/ssetools/${VER}" --exclude .git

# Replace the dynamic version script with the release version number
echo "__version__ = '$VER'" > "/software/sse2/generic/manual/ssetools/${VER}/mpprun/mpprun/_version.py"

# Create hpc module
# Note: the lmod file starts with setting HPC_SSE_VER, but it cannot use this variable internally, the versioned parts of the PATH must be hardcoded,
# since when unloading the module the variable gets unset first, which makes the module not unloading properly.
[ -d "/software/sse2/generic/modules/hpc/" ] || mkdir -p /software/sse2/generic/modules/hpc/
cat > "/software/sse2/generic/modules/hpc/.${VER_BARE}.lua" <<EOF
help([==[

Description
===========
This module sets up the HPC default user environment.
]==])

whatis([==[Description: default HPC user environment]==])

conflict("nsc")
conflict("hpc")

setenv("HPC_SSE_VER","${VER}")

prepend_path("PATH", "/software/tools/bin")
prepend_path("PATH", "/software/sse2/generic/manual/ssetools/${VER}/bin")
prepend_path{"PATH", "/software/sse2/generic/manual/ssetools/${VER}/wrappers",priority=2000}
prepend_path("MANPATH", "/usr/share/man")
setenv("HPC_LD_FLAG","1")

if userInGroups("nsc") then
   prepend_path("PATH", "/software/sse2/generic/manual/ssetools/${VER}/sbin")
end

-- Set standardized SNIC environment variables
setenv("SNIC_BACKUP",os.getenv("HOME"))
-- setenv("SNIC_NOBACKUP","only-set-in-job-environment")
-- setenv("SNIC_TMP,"only-set-in-job-environment")
setenv("SNIC_SITE","nsc")
setenv("SNIC_RESOURCE",os.getenv("NSC_RESOURCE_NAME"))

-- Avoid putting cached Singularity data in the default location (\$HOME)
-- Current versions of apptainer warns if SINGULARITY_CACHEDIR is set but
-- APPTAINER_CACHEDIR isn't, so set both. SINGULARITY_CACHEDIR should
-- eventually be removed.
setenv("SINGULARITY_CACHEDIR","/scratch/local")
setenv("APPTAINER_CACHEDIR","/scratch/local")

-- Configure srun, etc. to use the right MPI
setenv("SLURM_MPI_TYPE","pmi2")

-- Configure PSM2 in order to avoid OpenMPI needlessly doling out copious
-- warning messages about PSM2_CUDA features we don't have
setenv("PSM2_CUDA", "0")

-- Make sure the motd script is only ever run once
if mode() == "load" and os.getenv("NSC_MOTD_DONE") ~= "1" then
  setenv("NSC_MOTD_DONE","1")
  execute{cmd="/software/sse2/generic/manual/ssetools/${VER}/sbin/hpc_motd.sh",modeA={"load"}}
end

-- Setup container metadata env variable
local containerpath = "/software/sse2/" .. os.getenv("HPC_SSE_SYSNAME") .. "/containers"
local containers = ""
local sep = ""
if isDir(containerpath) then
  for filename in lfs.dir(containerpath) do
    local path = containerpath .. "/" .. filename
    if lfs.symlinkattributes(path,"mode") == "link" and not filename:match ".meta$|~$" then
	local target = lfs.symlinkattributes(path, "target")
	local max = 10
	while lfs.symlinkattributes(target,"mode") == "link" and max > 0 do
          target = lfs.symlinkattributes(target, "target")
          max = max - 1
	end

	local desc = "(no description)"
        local f = io.open(target .. ".meta", "r")
        if f~=nil then
          m = f:read "*a":match("\ndescription: ([^\n]*)")
          if m~=nil then
            desc = m
          end
          f:close()
	else
	  f = io.open(path .. ".meta", "r")
          if f~=nil then
            m = f:read "*a":match("\ndescription: ([^\n]*)")
            if m~=nil then
              desc = m
            end
            f:close()
          end
        end

        containers = containers .. sep .. filename:gsub("%..*$","") .. "," .. filename:gsub(".*%.","")  .. "," .. target .. "," .. desc
        sep = ":"
    end
  end
end
setenv("HPC_CONTAINERS",containers)

-- Old container metadata env variable
if isFile ("/software/admin/containers/images/centos7_compute_node.sif") then
  setenv("HPC_CONTAINERS_TETRALITH_EL7",capture("readlink -f /software/admin/containers/images/centos7_compute_node.sif"):gsub("\\n\$",""))
end

set_alias("usereservation", "source /software/sse2/generic/manual/ssetools/" .. os.getenv("HPC_SSE_VER") .."/bin/hpc_usereservation")
set_alias("hpc_usereservation", "source /software/sse2/generic/manual/ssetools/${VER}/bin/hpc_usereservation")

add_property("lmod","sticky")

EOF

echo "== ssetools ${VER} installed as: /software/sse2/generic/manual/ssetools/${VER}"
echo ""
echo "== hpc module created at: /software/sse2/generic/modules/hpc/.${VER_BARE}.lua"
echo ""
echo "Test the new module on a normal login node by loading it this way:"
echo ""
echo "  module --force purge"
echo "  module add hpc/.${VER_BARE}.lua"
echo ""
echo "After testing, if you have installed the latest version and want to make it the default hpc module, do:"
echo ""
echo "    ln -sfn \".${VER_BARE}.lua\" \"/software/sse2/generic/modules/hpc/default\""
echo ""
