--
-- This file is part of ssetools
-- Copyright © 2023-2024 The ssetools AUTHORS
--
-- Use of this source code is governed by a MIT-style license that can be
-- found in the LICENSE file of the ssetools repository.
--
-- History
--   rar@nsc: 2022 initial version based on a few old routines from NSC of unknown origin
--

require("strict")
local hook = require("Hook")

function split(inputstr, sep)
        local t={}
        for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
                table.insert(t, str)
        end
        return t
end

function myMsgHook(kind,a)
  if (kind == "spider") then
     table.insert(a,1,"####################################################################################################################################\n")
     table.insert(a,2,"# NOTE: At NSC the output of 'module spider' is generally not helpful as all relevant software modules are shown by 'module avail' #\n")
     table.insert(a,3,"# Some HPC centers hide software until the necessary dependencies have been loaded. NSC does not do that.                          #\n")
     table.insert(a,4,"####################################################################################################################################\n")
  end
  return a
end
hook.register("msgHook", myMsgHook)

function myIsVisibleHook(modT)
  local sysid = split(os.getenv("HPC_SSE_SYSID"),"-")
  -- io.stderr:write("B :: ",modT["fullName"]," :: ","(:"..sysid[1].."-"..sysid[2]..")$"," :: ",tostring(string.find(modT["fullName"],":("..sysid[1].."-"..sysid[2]..")$"))," :: ",tostring(string.find("hello_world/1.0:el7-x8664","(:el7[-]x8664)$")), "\n")
  if string.find(modT["fullName"],":",1,true) then
    if not (string.find(modT["fullName"],"generic") or string.find(modT["fullName"],"(:"..sysid[1]..")$") or string.find(modT["fullName"],"(:"..sysid[1].."[-]"..sysid[2]..")$") or string.find(modT["fullName"],"(:"..sysid[1].."[-]"..sysid[2].."[-]"..sysid[3]..")$")) then
      modT["isVisible"] = false
    end
  end
end
-- Skip this for now, we decided not to do a common module directory for all systems
-- hook.register("isVisibleHook", myIsVisibleHook)

function nsc_message_intel_buildenv()
  if mode() == 'load' and os.getenv("NSC_MODULE_SILENT") ~= "1" then
    LmodMessage('***************************************************')
    LmodMessage('You have loaded an intel buildenv module')
    LmodMessage('***************************************************')
    LmodMessage('The buldenv-intel module makes available:')
    LmodMessage(' - Compilers: icc, ifort, etc.')
    LmodMessage(' - Mpi library with mpi-wrapped compilers: intel mpi with mpiicc, mpiifort, etc.')
    LmodMessage(' - Numerical libraries: intel MKL')
    LmodMessage('')
    LmodMessage('It also makes a set of dependency library modules available via')
    LmodMessage('the regular module command. Just do:')
    LmodMessage('  module avail')
    LmodMessage('to see what is available.')
    LmodMessage('')
    LmodMessage('NOTE: You shoud never load build environments inside submitted jobs.')
    LmodMessage('(with the single exception of when using supercomputer time to compile code.)')
    LmodMessage('')
  end 
end

sandbox_registration{ nsc_message_intel_buildenv = nsc_message_intel_buildenv }

function nsc_message_gcc_buildenv()
  if mode() == 'load' and os.getenv("NSC_MODULE_SILENT") ~= "1" then
    LmodMessage('***************************************************')
    LmodMessage('You have loaded an gcc buildenv module')
    LmodMessage('***************************************************')
    LmodMessage('The buildenv-gcc module makes available:')
    LmodMessage(' - Compilers: gcc, gfortran, etc.')
    LmodMessage(' - MPI library with mpi-wrapped compilers: OpenMPI with mpicc, mpifort, etc.')
    LmodMessage(' - Numerical libraries: OpenBLAS, LAPACK, ScaLAPACK, FFTW')
    LmodMessage('')
    LmodMessage('It also makes a set of dependency library modules available via')
    LmodMessage('the regular module command. Just do:')
    LmodMessage('  module avail')
    LmodMessage('to see what is available.')
    LmodMessage('')
    LmodMessage('NOTE: You shoud never load build environments inside submitted jobs.')
    LmodMessage('(with the single exception of when using supercomputer time to compile code.)')
    LmodMessage('')
  end 
end

sandbox_registration{ nsc_message_gcc_buildenv = nsc_message_gcc_buildenv }

function nsc_message_on_load(message)
  if mode() == 'load' and os.getenv("NSC_MODULE_SILENT") ~= "1" then
    LmodMessage(message)
  end
end

sandbox_registration{ nsc_message_on_load = nsc_message_on_load }

function nsc_message_recommend(version)

  version = version or ""

  if (version == "" ) then
    LmodMessage('*************** NO MODULE LOADED *****************')
    LmodMessage('***** Please also specify desired version ********')
    LmodMessage('')
    LmodMessage('')
    LmodMessage('This is a notice to inform you that NSC has deprecated the use')
    LmodMessage('of default modules for this software. You will now instead need')
    LmodMessage('to specify version number/name as well.')
    LmodMessage('')
    LmodMessage('There is no current version recommendation for this software.')
    LmodMessage('')
    LmodError('No module loaded.')
  else
    LmodMessage('*************** NO MODULE LOADED *****************')
    LmodMessage('***** Please also specify desired version ********')
    LmodMessage('')
    LmodMessage('')
    LmodMessage('This is a notice to inform you that NSC has deprecated the use')
    LmodMessage('of default modules for this software. You will now instead need')
    LmodMessage('to specify version number/name as well.')
    LmodMessage('')
    LmodMessage('The recommended version of '.. myModuleName() ..' is ' .. version .. ', i.e., add it with')
    LmodMessage('')
    LmodMessage('  module load ' .. myModuleName() .. '/' .. version)
    LmodMessage('')
    LmodError('No module loaded.')
  end
end

sandbox_registration{ nsc_message_recommend = nsc_message_recommend }

function nsc_message_onload(message)
  if mode() == 'load' and os.getenv("NSC_MODULE_SILENT") ~= "1" then
    LmodMessage(message)
  end 
end

sandbox_registration{ nsc_message_onload = nsc_message_onload }

function nsc_message_needs_python27()
  if mode() == 'load' and os.getenv("NSC_MODULE_SILENT") ~= "1" then
     LmodMessage([==[
This module requires loading a suitable Python 2.7 environment alongside it, e.g.:

  module load Python/2.7.14-anaconda-5.0.1-nsc1
]==])
  end 
end

sandbox_registration{ nsc_message_needs_python27 = nsc_message_needs_python27 }


function nsc_message_needs_python3()
  if mode() == 'load' and os.getenv("NSC_MODULE_SILENT") ~= "1" then
     LmodMessage([==[
This module requires loading a suitable Python 3 environment alongside it, e.g.:

  module load Python/3.6.3-anaconda-5.0.1-nsc1
]==])
  end 
end

sandbox_registration{ nsc_message_needs_python3 = nsc_message_needs_python3 }


function nsc_message_needs_python27_or_3()
  if mode() == 'load' and os.getenv("NSC_MODULE_SILENT") ~= "1" then
     LmodMessage([==[
This module requires loading a suitable Python 2.7 or 3 environment alongside it, e.g.:

  module load Python/3.6.3-anaconda-5.0.1-nsc1

or

  module load Python/2.7.14-anaconda-5.0.1-nsc1
]==])
  end 
end

sandbox_registration{ nsc_message_needs_python27_or_3 = nsc_message_needs_python27_or_3 }




function nsc_set_mpivars(mod) 
  if mod == "intel" or mod == "icc" or mod == "ifort" then
	setenv("I_MPI_F77","ifort")
	setenv("I_MPI_F90","ifort")
	setenv("I_MPI_FC","ifort")
	setenv("I_MPI_CC","icc")
	setenv("I_MPI_CXX","icpc")
	setenv("OMPI_FC","ifort")
	setenv("OMPI_F90","ifort")
	setenv("OMPI_F77","ifort")
	setenv("OMPI_CC","icc")
	setenv("OMPI_CXX","icpc")
	setenv("OMPI_MPIF90","ifort")
	setenv("OMPI_MPIF77","ifort")
	setenv("OMPI_MPICC","icc")
	setenv("OMPI_MPICXX","icpc")
	setenv("OMPI_MPIXX","icpc")
	setenv("MPICH_F77","ifort")
	setenv("MPICH_F90","ifort")
	setenv("MPICH_CC","icc")
	setenv("MPICH_CXX","icpc")
  elseif mod == "gcc" or mod == "gfortran" then
  	setenv("I_MPI_F77","gfortran")
	setenv("I_MPI_F90","gfortran")
	setenv("I_MPI_FC","gfortran")
	setenv("I_MPI_CC","gcc")
	setenv("I_MPI_CXX","g++")
	setenv("OMPI_FC","gfortran")
	setenv("OMPI_F77","gfortran")
	setenv("OMPI_F90","gfortran")
	setenv("OMPI_CC","gcc")
	setenv("OMPI_CXX","g++")
	setenv("OMPI_MPIF77","gfortran")
	setenv("OMPI_MPIF90","gfortran")
	setenv("OMPI_MPICC","gcc")
	setenv("OMPI_MPICXX","g++")
	setenv("OMPI_MPIXX","g++")
	setenv("MPICH_F77","gfortran")
	setenv("MPICH_F90","gfortran")
	setenv("MPICH_CC","gcc")
	setenv("MPICH_CXX","g++")
  elseif mod == "clang" then 

  end
end
sandbox_registration{ nsc_set_mpivars = nsc_set_mpivars }

-- This function check that you are part of any one of a list of groups. The argument is a space separated string.
function nsc_require_group(groups)
     if (mode() == "load") then
         local grps  = capture("groups")
         local found = false
         local auth_group = ''
         for g in grps:split("[ \n]") do
             for auth_group in groups:split(",") do
                 if (g == auth_group)  then
                     found = true
                     break
                 end
             end
             if (found) then
                 break
             end
         end
         if (not found) then
             LmodError("\nYou do not have access to load this module. To get access, contact support@nsc.liu.se")
         end
     end
end
sandbox_registration{ nsc_require_group = nsc_require_group }

-- This helper function returns a Lua table with all modules that are presently loaded
function nsc_loaded_modules()
   local a = mcp:loaded_modules()
   return a
end
sandbox_registration{ nsc_loaded_modules = nsc_loaded_modules }

-- nsc_interpreter_version_check
--  // Rickard Armiento, Nov 2018 
--
-- This is a general function that checks which version of an interpreter is loaded and takes one of three actions:
--
--  - If it is the recommended version, or another whitelisted version: do nothing.
--  - If it is a blacklisted version: break and report an error.
--  - If it is an unrecoginzed version: print out a notice and hint about how to swap the version for the recommended one.
-- 
-- Usage: nsc_interpreter_version_check(category, recommended, whitelist, blacklist, version_reqs, required_patterns, blacklist_patterns)
--
--   All arguments except category and recommended are optional (leave them out, set them to nil, or set them to the empty list.)
--
--     * category = the name of the interpreter module, e.g., 'Python', 'Perl', 'Matlab', etc.
--
--     * recommended = the recommended version of the interpreter module, e.g., '3.6.4-nsc2-intel-2018a-eb'
--
--     * whitelist = a list of module versions to whitelist, e.g., {'3.6.3-anaconda-5.0.1-nsc1','2.7.14-anaconda-5.0.1-nsc1'}
--         Note: the recommended module is automatically regarded as whitelisted.
--               the matching is for the full exact version.
--
--     * blacklist = a list of module versions to blacklist, e.g., {'3.6.4-nsc1-intel-2018a-eb','2.7.14-nsc1-gcc-2018a-eb'}
--         Note: the matching is for the full exact version.
--
--     * version_reqs = a list of version intervals. The loaded module version must match any one of these, otherwise it is regarded as blacklisted
--         The basic format is: {{{from_major1,from_minor1},{to_major1, to_major1}}, {{from_major2,from_minor2},{to_major2,to_minor2}}, ...} 
--         There is considerable freedom in the format:
--           - Any minor or major entry can be nil, which means any value.
--           - Minor entries can be omitted, which means the same as setting them to nil.
--           - A whole {from_major,from_minor} segment can be nil, which means 'from any version' (i.e., version zero and up)
--           - A whole {to_major, to_minor} segment can be nil, which means 'to any version' (i.e., not limited from above)
--           - Instead of {{from_major1,from_minor1},{to_major1, to_major1}} one can just put {minor, major}, which means 'this version or higher'.
--         For example: { { {2,6},{2,nil} } , {3,7} }  means the version must either be within 2.6 to 2.(anything) OR within 3.7 to anything.
--         See other examples of possible setting below.
--
--     * required_patterns = if the module version is not in the whitelist or blacklist, it is must match at least one of the
--                          patterns in required_patterns. If it does not, it is regarded as blacklisted (i.e., an error is reported.)
--
--     * blacklist_patterns = if the module version matches any pattern given in blacklist_patterns, it is regarded as blacklisted
--                          (i.e., an error is reported.)
--
--   Examples: 
--
--     We have a module tested using Python/2.7.15-extras-nsc1-gcc-2018a-eb, and we know it is supposed to work with 
--     just about any version of Python, i.e., both 2 and 3. We don't know anything else about its requirements:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb")
--
--     The module is now also found working with Python/2.7.15-anaconda-5.0.1-nsc1 and Python/2.7.15-anaconda-5.0.2-nsc1:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb", {"2.7.15-anaconda-5.0.1-nsc1", "2.7.15-anaconda-5.0.2-nsc1"})
--
--     The module is found NOT to work with Python/3.6.4-nsc1-intel-2018a-eb:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb", {"2.7.15-anaconda-5.0.1-nsc1", "2.7.15-anaconda-5.0.2-nsc1"}, {"3.6.4-nsc1-intel-2018a-eb"})
--
--     For simplicty, lets set the whitelist and blacklist to nil from now on (but they could have been kept)
--     We find out that the module needs at least Python version 2.7:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb", nil, nil, {{2,7}})
--     
--     No, wait, it doesn't work with Python 3 at all:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb", nil, nil, {{{2,7},{2,nil}}})
-- 
--     Actually, it only works for exactly version 2.7:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb", nil, nil, {{{2,7},{2,7}}})
--
--     We find out that it works for either Python 2.7 OR Python 3.7 and onwards:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb", nil, nil, {{{2,7},{2,7}},{3,7}})
--
--     But, it needs a Python version with '-extras-' in the name:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb", nil, nil, {{{2,7},{2,7}},{3,7}},{'%-extras%-'})
--
--     Oh, and it not compatible with Python versions with '-bare-' in the name:
--
--       interpreter_version_check("Python","2.7.15-extras-nsc1-gcc-2018a-eb", nil, nil, {{{2,7},{2,7}},{3,7}},{'%-extras%-'},{'%-bare%-'})
--

-- (nsc_interpreter_blacklist_error is just a helper function for nsc_interpreter_version_check below)
--
function nsc_interpreter_blacklist_error(category, recommended)
	     LmodMessage("\n******************************************************************************************************")	       
	     LmodMessage("ERROR: You have a version of " .. category .. " loaded that is known to be")
	     LmodMessage("incompatible with the module you are trying to load: " .. myModuleFullName())	       
	     LmodMessage("\nYou can replace your " .. category .. " module with the recommended one by using the following commands:\n")
	     LmodMessage("   module unload " .. category)	        
	     LmodMessage("   module load " .. category .. "/" .. recommended .. "\n")	        
	     LmodMessage("******************************************************************************************************")	       
	     LmodError("Incompatible " .. category .. " module detected.")
end

function nsc_interpreter_version_check(category, recommended, whitelist, blacklist, version_reqs, required_patterns, blacklist_patterns)

     if os.getenv("NSC_MODULE_IGNORE_" .. string.upper(category) .. "_CHECK") == "1" or mode() ~= "load" then
     	return
     end

     if isloaded(category) then

        if isloaded(category .. "/" .. recommended) then	
	   return
	end

        if whitelist then
          for _,v in pairs(whitelist) do
            if isloaded(category .. "/" .. v) then
                return
            end
          end
        end
 
        if blacklist then
          for _,v in pairs(blacklist) do
            if isloaded(category .. "/" .. v) then
	        nsc_interpreter_blacklist_error(category, recommended)
    	        return
  	    end
          end
        end 

        local active = nsc_loaded_modules()
        local loaded_fullname = nil
	local loaded_version = nil
  
        -- DEBUG: dump contents of loaded_modules table
        --for index, data in ipairs(active) do
        --	 LmodMessage("" .. index)
        --
        --	 for key, value in pairs(data) do
        --      LmodMessage('\t' .. key .. "=" .. value)
        --	 end
        --end

        for i = 1,#active do	
           if (active[i].sn == category) then
                loaded_fullname = active[i].fullName
                loaded_version = string.sub(loaded_fullname,#category+2)
           end
	end

	if loaded_version and version_reqs then
	   local any_version_match = 0
	   local loaded_major = string.match(loaded_version,"(%d+)")	   
	   if loaded_major then
	      loaded_major = tonumber(loaded_major)
	   end
           local loaded_minor = string.match(loaded_version,"%d+%.(%d+)%.")	
	   if loaded_minor then
	      loaded_minor = tonumber(loaded_minor)
	   end
	   if loaded_major then
  	     for _,v in pairs(version_reqs) do
	       local this_version_match = 1

               local from_ver = nil
	       local to_ver = nil
               local from_major = nil
               local to_major = nil
               local from_minor = nil
               local to_minor = nil

	       if type(v[1]) == "table" then
	          from_ver = v[1]
		  to_ver = v[2]
               else
	          from_ver = v		  
	       end

	       if from_ver and from_ver[1] ~= nil then
                  from_major = from_ver[1]
		  if loaded_minor and #from_ver > 1 then
		    from_minor = from_ver[2]
		  end
	       end

	       if to_ver and to_ver[1] ~= nil then
                  to_major = to_ver[1]
		  if loaded_minor and #to_ver > 1 then
		    to_minor = to_ver[2]
		  end
	       end

	       if from_major and loaded_major < from_major then
	          this_version_match = 0
	       end
	       if from_major and loaded_major == from_major then
		   if loaded_minor and from_minor and loaded_minor < from_minor then
		     this_version_match = 0
                   end		 	      
	       end 

	       if to_major and loaded_major > to_major then
	         this_version_match = 0
	       end
	       if to_major and loaded_major == to_major then
		   if loaded_minor and to_minor and loaded_minor > to_minor then
                      this_version_match = 0
                   end		 	      
	       end
	       if this_version_match == 1 then 
                 any_version_match = 1
                 break
	       end
	     end
	     if any_version_match == 0 then
  	       nsc_interpreter_blacklist_error(category, recommended)
	       return
	     end
	   end
        end  

        -- For possible future use, but for now I don't think we should allow a whitelist pattern, either
        -- you have tested a version of the interpreter, or you have not.
        --
	-- if whitelist_pattern then
  	--  for _,v in pairs(whitelist_pattern) do
        --    if string.match(loaded_version,v) then
	--       return
        --    end
  	--  end
        -- end

        if required_patterns and #required_patterns > 0 then
  	  local required_match = 0
	  for _,v in pairs(required_patterns) do
            if string.match(loaded_version,v) then
	        required_match = 1
            end
  	  end
	  if required_match == 0 then
	    nsc_interpreter_blacklist_error(category, recommended)
	  end
        end

        if blacklist_patterns then
	  for _,v in pairs(blacklist_patterns) do
            if string.match(loaded_version,v) then
	       nsc_interpreter_blacklist_error(category, recommended)
	       return
            end
  	  end
        end

	LmodMessage("\nNOTICE: You have a version of " .. category .. " loaded which has not been tested if it works with the module being loaded: " .. myModuleFullName() .. "\n")	       
	LmodMessage("If you have trouble getting this software to work, try replacing your " .. category .. " module with the recommended one by using the following commands:\n")
	LmodMessage("   module unload " .. category)	        
	LmodMessage("   module load " .. category .. "/" .. recommended .. "\n")	        
	return
     end
     depends_on(category .. "/" .. recommended)
end

sandbox_registration{ nsc_interpreter_version_check = nsc_interpreter_version_check }

function nsc_group_or_message(groups, string)
    if (mode() == "load") then
        local grps  = capture("groups")
        local found = false
        local auth_group = ''
        for g in grps:split("[ \n]") do
            for auth_group in groups:split(",") do
                if (g == auth_group)  then
                    found = true
                    break
                end
            end
            if (found) then
                break
            end
        end
        if (not found) then
            LmodMessage(string)
        end
    end
end

sandbox_registration{ nsc_group_or_message = nsc_group_or_message }



-------------------------------------------------------------------------------
-- tinyyaml - YAML subset parser
-- From: https://github.com/peposso/lua-tinyyaml
--
-- MIT License
--
-- Copyright (c) 2017 peposso
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in all
-- copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
-------------------------------------------------------------------------------


local schar = string.char
local ssub, gsub = string.sub, string.gsub
local sfind, smatch = string.find, string.match
local tinsert, tremove = table.insert, table.remove

local UNESCAPES = {
  ['0'] = "\x00", z = "\x00", N    = "\x85",
  a = "\x07",     b = "\x08", t    = "\x09",
  n = "\x0a",     v = "\x0b", f    = "\x0c",
  r = "\x0d",     e = "\x1b", ['\\'] = '\\',
};

-------------------------------------------------------------------------------
-- utils
local function select(list, pred)
  local selected = {}
  for i = 0, #list do
    local v = list[i]
    if v and pred(v, i) then
      tinsert(selected, v)
    end
  end
  return selected
end

local function startswith(haystack, needle)
  -- io.stderr:write("------")
  -- io.stderr:write("" .. debug.traceback())
  -- io.stderr:write("------")
  return ssub(haystack, 1, #needle) == needle
end

local function ltrim(str)
  return smatch(str, "^%s*(.-)$")
end

local function rtrim(str)
  return smatch(str, "^(.-)%s*$")
end

-------------------------------------------------------------------------------
-- Implementation.
--
local class = {__meta={}}
function class.__meta.__call(cls, ...)
  local self = setmetatable({}, cls)
  if cls.__init then
    cls.__init(self, ...)
  end
  return self
end

function class.def(base, type, cls)
  base = base or class
  local mt = {__metatable=base, __index=base}
  for k, v in pairs(base.__meta) do mt[k] = v end
  cls = setmetatable(cls or {}, mt)
  cls.__index = cls
  cls.__metatable = cls
  cls.__type = type
  cls.__meta = mt
  return cls
end


local types = {
  null = class:def('null'),
  map = class:def('map'),
  omap = class:def('omap'),
  pairs = class:def('pairs'),
  set = class:def('set'),
  seq = class:def('seq'),
  timestamp = class:def('timestamp'),
}

local Null = types.null
function Null.__tostring() return 'yaml.null' end
function Null.isnull(v)
  if v == nil then return true end
  if type(v) == 'table' and getmetatable(v) == Null then return true end
  return false
end
local null = Null()

function types.timestamp:__init(y, m, d, h, i, s, f, z)
  self.year = tonumber(y)
  self.month = tonumber(m)
  self.day = tonumber(d)
  self.hour = tonumber(h or 0)
  self.minute = tonumber(i or 0)
  self.second = tonumber(s or 0)
  if type(f) == 'string' and sfind(f, '^%d+$') then
    self.fraction = tonumber(f) * math.pow(10, 3 - #f)
  elseif f then
    self.fraction = f
  else
    self.fraction = 0
  end
  self.timezone = z
end

function types.timestamp:__tostring()
  return string.format(
    '%04d-%02d-%02dT%02d:%02d:%02d.%03d%s',
    self.year, self.month, self.day,
    self.hour, self.minute, self.second, self.fraction,
    self:gettz())
end

function types.timestamp:gettz()
  if not self.timezone then
    return ''
  end
  if self.timezone == 0 then
    return 'Z'
  end
  local sign = self.timezone > 0
  local z = sign and self.timezone or -self.timezone
  local zh = math.floor(z)
  local zi = (z - zh) * 60
  return string.format(
    '%s%02d:%02d', sign and '+' or '-', zh, zi)
end


local function countindent(line)
  local _, j = sfind(line, '^%s+')
  if not j then
    return 0, line
  end
  return j, ssub(line, j+1)
end

local function parsestring(line, stopper)
  stopper = stopper or ''
  local q = ssub(line, 1, 1)
  if q == ' ' or q == '\t' then
    return parsestring(ssub(line, 2))
  end
  if q == "'" then
    local i = sfind(line, "'", 2, true)
    if not i then
      return nil, line
    end
    return ssub(line, 2, i-1), ssub(line, i+1)
  end
  if q == '"' then
    local i, buf = 2, ''
    while i < #line do
      local c = ssub(line, i, i)
      if c == '\\' then
        local n = ssub(line, i+1, i+1)
        if UNESCAPES[n] ~= nil then
          buf = buf..UNESCAPES[n]
        elseif n == 'x' then
          local h = ssub(i+2,i+3)
          if sfind(h, '^[0-9a-fA-F]$') then
            buf = buf..schar(tonumber(h, 16))
            i = i + 2
          else
            buf = buf..'x'
          end
        else
          buf = buf..n
        end
        i = i + 1
      elseif c == q then
        break
      else
        buf = buf..c
      end
      i = i + 1
    end
    return buf, ssub(line, i+1)
  end
  if q == '-' or q == ':' then
    if ssub(line, 2, 2) == ' ' or #line == 1 then
      return nil, line
    end
  end
  local buf = ''
  while #line > 0 do
    local c = ssub(line, 1, 1)
    if sfind(stopper, c, 1, true) then
      break
    elseif c == ':' and (ssub(line, 2, 2) == ' ' or #line == 1) then
      break
    elseif c == '#' and (ssub(buf, #buf, #buf) == ' ') then
      break
    else
      buf = buf..c
    end
    line = ssub(line, 2)
  end
  return rtrim(buf), line
end

local function isemptyline(line)
  return line == '' or sfind(line, '^%s*$') or sfind(line, '^%s*#')
end

local function startswithline(line, needle)
  return startswith(line, needle) and isemptyline(ssub(line, #needle+1))
end

local function parseflowstyle(line, lines)
  local stack = {}
  while true do
    if #line == 0 then
      if #lines == 0 then
        break
      else
        line = tremove(lines, 1)
      end
    end
    local c = ssub(line, 1, 1)
    if c == '#' then
      line = ''
    elseif c == ' ' or c == '\t' or c == '\r' or c == '\n' then
      line = ssub(line, 2)
    elseif c == '{' or c == '[' then
      tinsert(stack, {v={},t=c})
      line = ssub(line, 2)
    elseif c == ':' then
      local s = tremove(stack)
      tinsert(stack, {v=s.v, t=':'})
      line = ssub(line, 2)
    elseif c == ',' then
      local value = tremove(stack)
      if value.t == ':' or value.t == '{' or value.t == '[' then error() end
      if stack[#stack].t == ':' then
        -- map
        local key = tremove(stack)
        stack[#stack].v[key.v] = value.v
      elseif stack[#stack].t == '{' then
        -- set
        stack[#stack].v[value.v] = true
      elseif stack[#stack].t == '[' then
        -- seq
        tinsert(stack[#stack].v, value.v)
      end
      line = ssub(line, 2)
    elseif c == '}' then
      if stack[#stack].t == '{' then
        if #stack == 1 then break end
        stack[#stack].t = '}'
        line = ssub(line, 2)
      else
        line = ','..line
      end
    elseif c == ']' then
      if stack[#stack].t == '[' then
        if #stack == 1 then break end
        stack[#stack].t = ']'
        line = ssub(line, 2)
      else
        line = ','..line
      end
    else
      local s, rest = parsestring(line, ',{}[]')
      if not s then
        error('invalid flowstyle line: '..line)
      end
      tinsert(stack, {v=s, t='s'})
      line = rest
    end
  end
  return stack[1].v, line
end

local function parseblockstylestring(line, lines, indent)
  if #lines == 0 then
    error("failed to find multi-line scalar content")
  end
  local s = {}
  local firstindent = -1
  local endline = -1
  for i = 1, #lines do
    local ln = lines[i]
    local idt = countindent(ln)
    if idt <= indent then
      break
    end
    if ln == '' then
      tinsert(s, '')
    else
      if firstindent == -1 then
        firstindent = idt
      elseif idt < firstindent then
        break
      end
      tinsert(s, ssub(ln, firstindent + 1))
    end
    endline = i
  end

  local striptrailing = true
  local sep = '\n'
  local newlineatend = true
  if line == '|' then
    striptrailing = true
    sep = '\n'
    newlineatend = true
  elseif line == '|+' then
    striptrailing = false
    sep = '\n'
    newlineatend = true
  elseif line == '|-' then
    striptrailing = true
    sep = '\n'
    newlineatend = false
  elseif line == '>' then
    striptrailing = true
    sep = ' '
    newlineatend = true
  elseif line == '>+' then
    striptrailing = false
    sep = ' '
    newlineatend = true
  elseif line == '>-' then
    striptrailing = true
    sep = ' '
    newlineatend = false
  else
    error('invalid blockstyle string:'..line)
  end
  local eonl = 0
  for i = #s, 1, -1 do
    if s[i] == '' then
      tremove(s, i)
      eonl = eonl + 1
    end
  end
  if striptrailing then
    eonl = 0
  end
  if newlineatend then
    eonl = eonl + 1
  end
  for i = endline, 1, -1 do
    tremove(lines, i)
  end
  return table.concat(s, sep)..string.rep('\n', eonl)
end

local function parsetimestamp(line)
  local _, p1, y, m, d = sfind(line, '^(%d%d%d%d)%-(%d%d)%-(%d%d)')
  if not p1 then
    return nil, line
  end
  if p1 == #line then
    return types.timestamp(y, m, d), ''
  end
  local _, p2, h, i, s = sfind(line, '^[Tt ](%d+):(%d+):(%d+)', p1+1)
  if not p2 then
    return types.timestamp(y, m, d), ssub(line, p1+1)
  end
  if p2 == #line then
    return types.timestamp(y, m, d, h, i, s), ''
  end
  local _, p3, f = sfind(line, '^%.(%d+)', p2+1)
  if not p3 then
    p3 = p2
    f = 0
  end
  local zc = ssub(line, p3+1, p3+1)
  local _, p4, zs, z = sfind(line, '^ ?([%+%-])(%d+)', p3+1)
  if p4 then
    z = tonumber(z)
    local _, p5, zi = sfind(line, '^:(%d+)', p4+1)
    if p5 then
      z = z + tonumber(zi) / 60
    end
    z = zs == '-' and -tonumber(z) or tonumber(z)
  elseif zc == 'Z' then
    p4 = p3 + 1
    z = 0
  else
    p4 = p3
    z = false
  end
  return types.timestamp(y, m, d, h, i, s, f, z), ssub(line, p4+1)
end

local function parsescalar(line, lines, indent)
  line = ltrim(line)
  line = gsub(line, '%s*#.*$', '')

  if line == '' or line == '~' then
    return null
  end

  local ts, _ = parsetimestamp(line)
  if ts then
    return ts
  end

  local s, _ = parsestring(line)
  if s and s ~= line then
    return s
  end

    -- Special cases
  if sfind('\'"!$', ssub(line, 1, 1), 1, true) then
    error('unsupported line: '..line)
  end

  if startswithline(line, '{}') then
    return {}
  end
  if startswithline(line, '[]') then
    return {}
  end

  if startswith(line, '{') or startswith(line, '[') then
    return parseflowstyle(line, lines)
  end

  if startswith(line, '|') or startswith(line, '>') then
    return parseblockstylestring(line, lines, indent)
  end

  -- Regular unquoted string
  local v = line
  if v == 'null' or v == 'Null' or v == 'NULL'then
    return null
  elseif v == 'true' or v == 'True' or v == 'TRUE' then
    return true
  elseif v == 'false' or v == 'False' or v == 'FALSE' then
    return false
  elseif v == '.inf' or v == '.Inf' or v == '.INF' then
    return math.huge
  elseif v == '+.inf' or v == '+.Inf' or v == '+.INF' then
    return math.huge
  elseif v == '-.inf' or v == '-.Inf' or v == '-.INF' then
    return -math.huge
  elseif v == '.nan' or v == '.NaN' or v == '.NAN' then
    return 0 / 0
  elseif sfind(v, '^[%+%-]?[0-9]+$') or sfind(v, '^[%+%-]?[0-9]+%.$')then
    return tonumber(v)  -- : int
  elseif sfind(v, '^[%+%-]?[0-9]+%.[0-9]+$') then
    return tonumber(v)
  end
  return v
end

local parsemap;  -- : func

local function parseseq(line, lines, indent)
  local seq = setmetatable({}, types.seq)
  if line ~= '' then
    error()
  end
  while #lines > 0 do
    -- Check for a new document
    line = lines[1]
    if startswith(line, '---') then
      while #lines > 0 and not startswith(lines, '---') do
        tremove(lines, 1)
      end
      return seq
    end

    -- Check the indent level
    local level = countindent(line)
    if level < indent then
      return seq
    elseif level > indent then
      error("found bad indenting in line: ".. line)
    end

    local i, j = sfind(line, '%-%s+')
    if not i then
      i, j = sfind(line, '%-$')
      if not i then
        return seq
      end
    end
    local rest = ssub(line, j+1)

    if sfind(rest, '^[^\'\"%s]*:') then
      -- Inline nested hash
      local indent2 = j
      lines[1] = string.rep(' ', indent2)..rest
      tinsert(seq, parsemap('', lines, indent2))
    elseif isemptyline(rest) then
      tremove(lines, 1)
      if #lines == 0 then
        tinsert(seq, null)
        return seq
      end
      if sfind(lines[1], '^%s*%-') then
        local nextline = lines[1]
        local indent2 = countindent(nextline)
        if indent2 == indent then
          -- Null seqay entry
          tinsert(seq, null)
        else
          tinsert(seq, parseseq('', lines, indent2))
        end
      else
        -- - # comment
        --   key: value
        local nextline = lines[1]
        local indent2 = countindent(nextline)
        tinsert(seq, parsemap('', lines, indent2))
      end
    elseif rest then
      -- Array entry with a value
      tremove(lines, 1)
      tinsert(seq, parsescalar(rest, lines))
    end
  end
  return seq
end

local function parseset(line, lines, indent)
  if not isemptyline(line) then
    error('not seq line: '..line)
  end
  local set = setmetatable({}, types.set)
  while #lines > 0 do
    -- Check for a new document
    line = lines[1]
    if startswith(line, '---') then
      while #lines > 0 and not startswith(lines, '---') do
        tremove(lines, 1)
      end
      return set
    end

    -- Check the indent level
    local level = countindent(line)
    if level < indent then
      return set
    elseif level > indent then
      error("found bad indenting in line: ".. line)
    end

    local i, j = sfind(line, '%?%s+')
    if not i then
      i, j = sfind(line, '%?$')
      if not i then
        return set
      end
    end
    local rest = ssub(line, j+1)

    if sfind(rest, '^[^\'\"%s]*:') then
      -- Inline nested hash
      local indent2 = j
      lines[1] = string.rep(' ', indent2)..rest
      set[parsemap('', lines, indent2)] = true
    elseif sfind(rest, '^%s+$') then
      tremove(lines, 1)
      if #lines == 0 then
        tinsert(set, null)
        return set
      end
      if sfind(lines[1], '^%s*%?') then
        local indent2 = countindent(lines[1])
        if indent2 == indent then
          -- Null array entry
          set[null] = true
        else
          set[parseseq('', lines, indent2)] = true
        end
      end

    elseif rest then
      tremove(lines, 1)
      set[parsescalar(rest, lines)] = true
    else
      error("failed to classify (2) line: "..line)
    end
  end
  return set
end

function parsemap(line, lines, indent)
  if not isemptyline(line) then
    error('not map line: '..line)
  end
  local map = setmetatable({}, types.map)
  while #lines > 0 do
    line = lines[1]
    -- Strip empty lines and comments
    while isemptyline(line) do
        tremove(lines, 1)
        if #lines == 0 then
	  return map
	end
        line = lines[1]
    end
    -- Check for a new document
    if startswith(line, '---') then
      while #lines > 0 and not startswith(lines, '---') do
        tremove(lines, 1)
      end
      return map
    end

    -- Check the indent level
    local level, _ = countindent(line)
    if level < indent then
      return map
    elseif level > indent then
      error("found bad indenting in line: ".. line)
    end

    -- Find the key
    local key
    local s, rest = parsestring(line)

    -- Quoted keys
    if s and startswith(rest, ':') then
      local sc = parsescalar(s, {}, 0)
      if sc and type(sc) ~= 'string' then
        key = sc
      else
        key = s
      end
      line = ssub(rest, 2)
    else
      error("failed to classify (1) line: "..line)
    end

    if map[key] ~= nil then
      error("found a duplicate key '"..key.."' in line: "..line)
      local suffix = 1
      while map[key..'__'..suffix] do
        suffix = suffix + 1
      end
      key = key ..'_'..suffix
    end

    line = ltrim(line)

    if ssub(line, 1, 1) == '!' then
      -- ignore type
      local rh = ltrim(ssub(line, 3))
      local typename = smatch(rh, '^!?[^%s]+')
      line = ltrim(ssub(rh, #typename+1))
    end

    if not isemptyline(line) then
      tremove(lines, 1)
      line = ltrim(line)
      map[key] = parsescalar(line, lines, indent)
    else
      -- An indent
      tremove(lines, 1)
      if #lines == 0 then
        map[key] = null
        return map;
      end
      if sfind(lines[1], '^%s*%-') then
        local indent2 = countindent(lines[1])
        map[key] = parseseq('', lines, indent2)
      elseif sfind(lines[1], '^%s*%?') then
        local indent2 = countindent(lines[1])
        map[key] = parseset('', lines, indent2)
      else
        local indent2 = countindent(lines[1])
        if indent >= indent2 then
          -- Null hash entry
          map[key] = null
        else
          map[key] = parsemap('', lines, indent2)
        end
      end
    end
  end
  return map
end


-- : (list<str>)->dict
local function parsedocuments(lines)
  -- lines = select(lines, function(s) return not isemptyline(s) end)

  if sfind(lines[1], '^%%YAML') then tremove(lines, 1) end

  local root = {}
  local in_document = false
  while #lines > 0 do
    local line = lines[1]
    -- Do we have a document header?
    local docright;
    if sfind(line, '^%-%-%-') then
      -- Handle scalar documents
      docright = ssub(line, 4)
      tremove(lines, 1)
      in_document = true
    end
    if docright then
      if (not sfind(docright, '^%s+$') and
          not sfind(docright, '^%s+#')) then
        tinsert(root, parsescalar(docright, lines))
      end
    elseif #lines == 0 or startswith(line, '---') then
      -- A naked document
      tinsert(root, null)
      while #lines > 0 and not sfind(lines[1], '---') do
        tremove(lines, 1)
      end
      in_document = false
    -- XXX The final '-+$' is to look for -- which ends up being an
    -- error later.
    elseif not in_document and #root > 0 then
      -- only the first document can be explicit
      error('parse error: '..line)
    -- Empty line?
    elseif isemptyline(line) then
      tremove(lines, 1)
    elseif sfind(line, '^%s*%-') then
      -- An array at the root
      tinsert(root, parseseq('', lines, 0))
    elseif sfind(line, '^%s*[^%s]') and not isemptyline(line) then
      -- A hash at the root
      local level = countindent(line)
      tinsert(root, parsemap('', lines, level))
    else
      -- Shouldn't get here.  @lines have whitespace-only lines
      -- stripped, and previous match is a line with any
      -- non-whitespace.  So this clause should only be reachable via
      -- a perlbug where \s is not symmetric with \S

      -- uncoverable statement
      error('parse error: '..line)
    end
  end
  if #root > 1 and Null.isnull(root[1]) then
    tremove(root, 1)
    return root
  end
  return root
end


local function parse(source)
  local lines = {}
  for line in string.gmatch(source..'\n', '(.-)\n') do
    tinsert(lines, line)
  end
  local docs = parsedocuments(lines)
  if #docs == 1 then
    return docs[1]
  end
  return docs
end

-------------------------------------------------------------------------------
-- END OF tinyyaml - YAML subset parser
-- From: https://github.com/peposso/lua-tinyyaml
-------------------------------------------------------------------------------

local function read_yaml_header(filename)

  local file = assert(io.open(filename,"r"))
  local line = file:read()

  -- Is there a yaml header?
  local header_spaces = string.match(line, "%-%-( +)%-%-%- *$")
  if header_spaces then
    local yaml = ""
    local prefix = "^%-%-" .. header_spaces
    for line in file:lines() do 
      if not string.match(line, "^%-%- *$") then
        if string.match(line, prefix) then
          if string.match(line, prefix .. "%-%-%- *$") or string.match(line, prefix .. "%.%.%. *$") then
            break
          end
          yaml = yaml .. string.sub(line,3+string.len(header_spaces)) .. "\n"
        else
          break
        end  
      end
    end
    file:close()
    -- io.stderr:write(yaml)
    if(yaml) then
      return parse(yaml)    
    else
      return {}
    end
  else
    file:close()
    return {}
  end

end

local function is_array(t)
  local i = 0
  for _ in pairs(t) do
    i = i + 1
    if t[i] == nil then return false end
  end
  return true
end

local function table_to_str(t)
  local vals = {}
  if(is_array(t)) then
    for k,v in pairs(t) do    
      if (type(v) == "table") then
        table.insert(vals, table_to_str(v))
      else
        table.insert(vals, tostring(v))
      end
    end
    return string.format("[ %s ]", table.concat(vals, ', ') )
  else
    for k,v in pairs(t) do    
      if (type(v) == "table") then
        table.insert(vals, k .. ": " .. table_to_str(v))
      else
        table.insert(vals, k .. ": " .. tostring(v))
      end
    end
    return string.format("{ %s }", table.concat(vals, ', ') )
  end
end


function nsc_setup_help()

	 local t = read_yaml_header(myFileName())

	 local description = ""
	 local more_info = ""
	 local nscdoc = ""
	 local support_tier = ""
	 local extra_help = ""

	 local header = "Help for " .. myModuleFullName() .. "\n"
	 header = string.rep("=",string.len(header)) .. "\n" .. header .. string.rep("=",string.len(header)) .. "\n\n"

         -- DESCRIPTION
 	 if (t["description"] ~= nil) then
	     description = [==[
Description
===========
]==] .. t["description"] .. "\n"
         else
	     description = [==[
Description
===========
This software appear to lack a description. If you get this message for normal user software, please
report this issue to: support@nsc.liu.se
]==]
         end

         -- NSC USER HELP
	 if (t["nscdoc"] ~= nil) then
	     nscdoc = [==[

NSC User help
=============
]==] .. "https://www.nsc.liu.se/software/installed/tetralith/" .. t["nscdoc"] .. "\n"
     	 else
	     nscdoc = [==[

NSC User help
=============
The software provided by this module appear not to have a general NSC user help web page. If
you have trouble, direct any questions to: support@nsc.liu.se
]==]
         end

         -- NSC SUPPORT TIER
	 if (t["support_tier"] == 1) then
	     support_tier = [==[

NSC Support level
=================
NSC has at least one application expert on staff with deep knowledge about this software, 
and we will try to help you with any problems. We have also done some internal tests to 
make sure that the program is working as intended. 
NSC also commits to installing updates as they become available.
]==]
         elseif (t["support_tier"] == 2) then
	     support_tier = [==[

NSC Support level
=================
NSC has only limited experience with this software, but we will try to help as much as possible. 
For example, if the program comes with a test suite, we will run it, and attempt to evaluate 
the results. We will try to install and test new versions, as soon as we can.
]==]
         elseif (t["support_tier"] == 3) then
	     support_tier = [==[

NSC Support level
=================
This is a test installation or custom software we helped install by request of a users. 
In general, we will not be able to help you much with this software. 
These types of installations can be completely untested, and will not be 
updated unless you send a request to NSC.
]==]
	 else
	     support_tier = ""
         end

         -- EXTRA HELP
	 if (t["help_extra"]) then
	     extra_help = "\n" .. t["help_extra"]
	 end

         -- HOMEPAGE
	 if (t["homepage"] ~= nil) then
	     more_info = [==[

More information
================
 - Homepage: ]==] .. t["homepage"] .. "\n"
         end

         if t["help_override"] then
	   help(t["help_override"])
         else
  	   help(header .. description .. nscdoc .. support_tier .. extra_help .. more_info)
         end
	 whatis("Description: " .. t["description"])
         whatis("Homepage: " .. t["homepage"])      
         whatis("NSC documentation: https://www.nsc.liu.se/software/installed/tetralith/" .. t["nscdoc"])      

         if (t["info"]) then
  	   for k, v in pairs(t["info"]) do
	     if (not (string.match(k,"_") or string.match(k,"x_"))) then
	       if (type(v) == "table") then
  	         whatis(k .. ": " .. table_to_str(v))
               else
	         whatis(k .. ": " .. tostring(v))	
               end
             end 
	   end
         end
end

sandbox_registration{ nsc_setup_help = nsc_setup_help }

