#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.

"""
This is a helper program to figure out what MPI launcher to use, with which arguments, in what environment (including which HPC modules to load) when launching a binary or script that uses MPI.
"""

import argparse, os, sys, logging, traceback, importlib, pkgutil, shutil, textwrap, shlex, syslog, re
from pprint import pformat

def module(action, *args):
    if 'HPC_MPPRUN_NO_MODULES' not in os.environ or os.environ['HPC_MPPRUN_NO_MODULES'] != '1':
        try:
            from env_modules_python import module as lmod_module
        except:
            logging.critical("Failed to import 'env_modules_python' to execute: module "+action+" "+" ".join(args))
        lmod_module(action,*args)
    else:
        logging.debug("Skipping module command due to env variable HPC_MPPRUN_NO_MODULES=1: module "+action+" "+" ".join(args))

from ._version import __version__
from .exceptionwrapper import ExceptionWrapper
from .hwinfo import Hwinfo

arguments = [
    {
        'names': ['--version'],
        'action': 'version',
        'version': '%(prog)s '+ __version__
    },
    {
        'names': ['executable'],
        'help': 'The binary or script to execute.',
        'type': str,
    },
    {
        'names': ['arguments'],
        'help': 'Arguments to pass to the executable.',
        'type': str,
        'nargs': argparse.REMAINDER,
    },
    {
        'names': ['-n', '--nranks', '--np'],
        'help': 'Specify the number of MPI tasks to run. (For compatibility, "-np" also works)',
        'type': int,
    },
    {
        'names': ['--launcher'],
        'help': 'Specify a preferred underlying MPI launcher to use. Use "native" for the launcher used to build the software with (if it can be determined).',
        'type': str,
    },
    {
        'names': ['--handler'],
        'help': 'Specify a preferred launcher handler to use',
        'type': str,
    },
    {
        'names': ['--no-smt', '--no-ht'],
        'help': 'On machines with simultaneous multithreading (also known as hyperthreading), use default values for cpus per task, threading, etc., that aim to place only one task per physical cpu core. ' \
                '(However, this may not hold if  one overrides the default values, e.g., by also explicitly setting cpus-per-task.)',
        'action': 'store_true',
        'default': False
    },
    {
        'names': ['-C', '--container', '--compat'],
        'help': 'Run executable inside the container / compatibility wrapper with the given name. For example, if there exists a container named el7 (e.g., to mimic an Enterprise Linux 7 system), then -C el7 will run the executable inside this wrapper.',
        'choices': [],
        'type': str,
        'metavar': '',
        'default': argparse.SUPPRESS,
    },
    {
        'names': ['--pass'],
        'help': 'Pass options to the underlying MPI launcher (also requires --launcher). Note: one MUST use the form --pass="--example" to handle usual arguments that starts with one or more dashes.',
        'dest': 'extra_launch_args',
        'action': 'append',
        'type': str,
        'default': []
    },
    {
        'names': ['--no-modules'],
        'help': 'Avoid all interactions with the module system. This option can also be enabled by setting the environment variable HPC_MPPRUN_NO_MODULES=1.',
        'action': 'store_true',
        'default': argparse.SUPPRESS,
    },
    {
        'names': ['-c', '--cpus-per-task'],
        'help': 'Configure how to use CPU cores and threads. The specified amount CPU cores is allocated per MPI process. ' \
                'The default is the value of the environment variable SLURM_CPUS_PER_TASK (or 1 if unset), which means that ' \
                'it is set by the SLURM job configuration (e.g., the "-c" flag to sbatch). ' \
                'Environment variables that control the number of threads, e.g., OMP_NUM_THREADS, if unset, will be set to the '\
                'same value (i.e., one OpenMP thread per core). ',
        'type': int,
    },
    {
        'names': ['-d', '--debug'],
        'help': 'Produce full tracebacks on error',
        'action': 'store_true',
        'default': False
    },
    {
        'names': ['-v', '--verbose'],
        'help': 'Increase verbosity of output',
        'dest': 'verbosity',
        'action': 'count',
        'default': 3
    },
    {
        'names': ['-q', '--quiet'],
        'help': 'Decrease verbosity of output',
        'action': 'count',
        'default': 0
    },
    {
        'names': ['-i', '--info'],
        'help': 'Only inspect the executable and show available launchers/options.',
        'action': 'store_true',
        'default': False
    },
    {
        'names': ['--allinfo'],
        'help': "Same as --info, but list *all* launchers (e.g., including those for other systems and compatibility layers).",
        'action': 'store_true',
        'default': False
    },

]

class NewlinePreservingArgumentDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def _split_lines(self, text, width):
        return [wrapped_line for line in text.splitlines() for wrapped_line in textwrap.wrap(re.sub(' +', ' ', line).rstrip(), width)]

def parse_args(argv):

    # Initialize argparser
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=NewlinePreservingArgumentDefaultsHelpFormatter)

    # Fill in dynamic options
    delayed_warnings = []
    containers = {}
    try:
        if 'HPC_CONTAINERS' in os.environ:
            containers = {y[0]:{'type':y[1], 'image':y[2], 'desc':y[3]} for y in [x.split(',') for x in os.environ['HPC_CONTAINERS'].split(':') if x.strip() != ""]}
    except Exception as e:
        delayed_warnings += ["Failed to parse HPC_CONTAINERS env variable: "+str(e)]

    container_arg = list(filter(lambda x:'--container' in x['names'], arguments))[0]
    if len(containers) > 0:
        container_arg['choices'] = containers.keys()
        container_arg['help'] += "\nContainers available in the current environement: \n"+"\n".join(["* "+key+": "+val['desc'] for key,val in containers.items()])
    else:
        container_arg['help'] += " (However, the current environment does not define any named containers.)"

    for arg in arguments:
        names = arg.pop('names')
        parser.add_argument(*names, **arg)

    # Handle the legacy '-np' option
    compat_argv = [(x if x!='-np' else '--np') for x in argv[1:]]

    # Detect attempts to call mpprun using the srun multibinary syntrax
    if ':' in argv:
        parser.error("It appears you are trying to use mpprun with a multibinary syntax (i.e., one of the arguments is a single colon ':'). This is not supported. Please instead launch your application with 'srun' and use the --multi-prog argument for a multibinary launch.")

    args = parser.parse_args(compat_argv)

    # Handle default None for suppressed default container
    if 'container' not in args:
        args.container = None

    # Make sure verbosity is in the allowed range
    log_levels = [logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
    args.log_level = log_levels[max(min(len(log_levels) - 1, args.verbosity-args.quiet),0)]

    # Configure logging
    logging.basicConfig(format='mpprun %(levelname)s: %(message)s',level=args.log_level)
    logging.debug("invoked with: "+str(args))

    # Turn on tracebacks, etc., if verbosity is max *or* the debug flag is given
    args.debug = args.debug or args.log_level <= logging.DEBUG
    ExceptionWrapper.debug = args.debug

    # Enforce that --pass requires --launcher
    if args.extra_launch_args != [] and args.launcher is None:
        parser.error("Use of --pass requires specifying an MPI launcher with --launcher.")

    # If SLURM_CPUS_PER_TASK it set, it is the default for -c
    if args.cpus_per_task is None:
        if 'SLURM_CPUS_PER_TASK' in os.environ:
            try:
                args.cpus_per_task = int(os.environ['SLURM_CPUS_PER_TASK'])
            except ValueError:
                args.cpus_per_task = None

    # Enforce that executable exists and is executable
    args.exec_path = shutil.which(args.executable)
    if args.exec_path is None and os.path.exists(os.path.join(".",args.executable)):
        args.exec_path = os.path.join(".",args.executable)
    if args.exec_path is None:
        parser.error("Specified executable '"+args.executable+"' does not exist or is not executable.")

    # Allinfo implies info
    if args.allinfo:
        args.info = True

    # info implies do_all so we get info from all launchers
    if args.info:
        args.do_all = True
    else:
        args.do_all = False

    # set env variable HPC_MPPRUN_NO_MODULES based on config
    if 'no_modules' in args and args.no_modules:
        os.environ['HPC_MPPRUN_NO_MODULES'] = '1'

    logging.debug("executable resolved to: "+str(args.exec_path))

    # output delayed warning
    for warning in delayed_warnings:
        logging.warning(warning)

    return args

def cli(args):

    try:

        logging.info("mpprun "+__version__+" started")

        try:
            logging.debug("discover hardware")
            hwinfo = Hwinfo()

            # Adjust defaults based on hwinfo
            if args.no_smt and hwinfo.smt_threads_per_physical_core is None:
                raise Exception("Unable to determine threads per physical core and the option --no-smt requested")

            if args.cpus_per_task is None and args.no_smt:
                args.cpus_per_task = hwinfo.smt_threads_per_physical_core

        except Exception as e:
            raise ExceptionWrapper("Error while discovering hardware", e) from e
        try:
            logging.debug("running handler discovery")

            discovered_handlers = [
                name for finder, name, ispkg
                in pkgutil.iter_modules()
                if name.startswith('mpprun_handler_')
            ]
            discovered_handlers.sort()

            logging.debug("available handlers: "+str(", ".join(discovered_handlers)))

            if len(discovered_handlers) == 0:
                raise Exception("mpprun internal error: no launcher handlers discovered (not even the default one)")

        except Exception as e:
            raise ExceptionWrapper("Error while disovering mpprun handlers", e) from e


        try:
            logging.debug("going through available handlers")

            handler_log_level=logging.DEBUG

            handlers_specs = []
            for handler in discovered_handlers:
                logging.debug("considering handler: "+str(handler))
                handler_name = '_'.join(handler.split('_')[3:])
                if args.handler is not None and args.handler != handler_name:
                    logging.debug("handler disregarded since --handler='"+args.handler+"' != "+handler_name)
                    continue
                try:
                    mod = importlib.import_module(handler)
                    handler_spec_list = mod.handle(args.exec_path, args.arguments, args, hwinfo)
                except Exception as e:
                    if args.debug:
                        raise
                    handler_spec_list = { 'viable': False, 'reason': 'internal error in handler (use -d for backtrace): '+str(e)}

                if not isinstance(handler_spec_list, list):
                    handler_spec_list = [handler_spec_list]

                for handler_spec in handler_spec_list:

                    handler_spec_defaults = {"handler": handler_name, "provider": mod.__file__, "viable": False, "launcher": 'unknown', "warnings": [],
                                             "info": [], "syslog": [], "native": False, "reason": None, "env": {}, "modules":[], "launch_args": [],
                                             "mpprun_default_args": {}}

                    if handler_spec is None:
                        handler_spec = handler_spec_defaults
                    else:
                        for key in handler_spec_defaults:
                            if key not in handler_spec:
                                handler_spec[key] = handler_spec_defaults[key]

                    # Check if the handler_spec specifies a specific container, but the user has specified a different container
                    if args.container is not None and (('compat' in handler_spec and handler_spec['compat'] not in [args.container, '*']) or ('container' in handler_spec and handler_spec['container'] not in [args.container, '*'])):
                        msg = "does not match specified container: "+str(args.container)
                        if args.allinfo:
                            logging.log(handler_log_level, "turn handler "+handler_name+" unviable because it "+msg)
                            handler_spec['viable'] = False
                            handler_spec['reason'] = msg
                        else:
                            logging.log(handler_log_level, "ignoring handler "+handler_name+" because it "+msg)
                            continue


                    if 'sysname' in handler_spec and handler_spec['sysname'] != hwinfo.sysname:
                        msg = "specifies a sysname that does not match the current system: "+str(args.container)
                        if args.allinfo:
                            logging.log(handler_log_level, "turn handler "+handler_name+" unviable because it "+msg)
                            handler_spec['viable'] = False
                            handler_spec['reason'] = msg
                        else:
                            logging.log(handler_log_level, "ignoring handler "+handler_name+" because it "+msg)
                            continue

                    if handler_spec['viable'] or args.info:
                        if args.launcher is not None and ((handler_spec['launcher'] != args.launcher) and (os.path.basename(handler_spec['launcher']) != args.launcher) and not (args.launcher == 'native' and handler_spec['native'])):
                            logging.debug("handler disregarded since: --lancher="+args.launcher+" != "+handler_spec['launcher'])
                            continue
                        handlers_specs += [handler_spec]

                    if handler_spec['viable']:
                        logging.log(handler_log_level, "handler "+handler_name+" found viable")
                    elif handler_spec['reason'] is not None:
                        logging.log(handler_log_level, "dismissed handler "+handler_name+": "+handler_spec['reason'])
                    else:
                        logging.log(handler_log_level, "dismissed handler "+handler_name+" (no reason given)")

                    if not args.do_all and len(handlers_specs) > 0:
                        break

            logging.debug("handlers found:"+str(handlers_specs))

        except Exception as e:
            raise ExceptionWrapper("Error while finding launcher to run",e) from e

        if not args.info:
            logging.debug("launch executable")

            try:
                if len(handlers_specs) == 0:
                    logging.critical("*************************************************************")
                    logging.critical("mpprun: cannot launch executable, no viable handler available")
                    logging.critical("(try mpprun -i for more info).")
                    logging.critical("*************************************************************")
                    return 1

                if "SLURM_JOB_ID" not in os.environ.keys() and "SLURM_JOBID" not in os.environ.keys():
                    logging.critical("mpprun requires to be run in a slurm allocation to launch an executable, and the SLURM_JOB* variables are not set.")
                    return 1

                handler_spec = handlers_specs[0]

                if 'mpprun_default_args' in handler_spec and isinstance(handler_spec['mpprun_default_args'], dict) and len(handler_spec['mpprun_default_args']) > 0:
                    logging.debug("processing mpprun_default_args from handler")
                    for ik, iv in handler_spec['mpprun_default_args'].items():
                        if not hasattr(args,ik) or getattr(args,ik) is None:
                            logging.debug("setting mpprun argument "+str(ik)+" to "+str(iv))
                            setattr(args,ik,iv)

                launch_manager = "mpprun_launcher_"+handler_spec["launcher"]
                if importlib.find_loader(launch_manager) is not None:
                    try:
                        logging.debug("invoking launch manager: "+launch_manager)
                        mod = importlib.import_module(launch_manager)
                        new_handler_spec = mod.prepare(args.exec_path, args.arguments, args, handler_spec, hwinfo)
                        handler_spec = new_handler_spec
                    except Exception as e:
                        if args.debug:
                            raise
                        else:
                            logging.warning("internal error in launch manager "+launch_manager+" (use -d for backtrace): "+str(e))
                else:
                    logging.debug("no launch manager found for launcher: "+handler_spec["launcher"])

                if 'warnings' in handler_spec:
                    for warn in handler_spec['warnings']:
                        logging.warning(warn)

                if 'info' in handler_spec:
                    for info in handler_spec['info']:
                        logging.info(str(info))

                # Silence help messages from Lmod
                # (We do this before configuring the environment so a handler can unset this
                # if that for some reason is desired)
                os.environ["HPC_MODULE_SILENT"]="1"
                os.environ["NSC_MODULE_SILENT"]="1"

                if 'HPC_MPPRUN_NO_MODULES' not in os.environ or os.environ['HPC_MPPRUN_NO_MODULES'] != '1':
                    if 'modules' in handler_spec:
                        for module_id in handler_spec['modules']:
                            if not module_id.startswith('!'):
                                module("-q","load",module_id)
                            elif module_id[1:] in [x.partition('/')[0] for x in os.environ['LOADEDMODULES'].split(':')]:
                                logging.info("Unloading module "+module_id[1:]+" (on request by handler)")
                                module("-q","unload",module_id[1:])

                if 'env' in handler_spec:
                    for key in handler_spec['env']:
                        if key.startswith("!") and key[1:] in os.environ:
                            logging.info("Unsetting env variable: "+str(key[:1]))
                            del os.environ[key[1:]]
                        elif key.startswith("~"):
                            if key not in os.environ:
                                logging.info("Setting env variable: "+str(key[:1])+"="+str(handler_spec['env'][key]))
                                os.environ[key[1:]] = handler_spec['env'][key]
                        else:
                            logging.info("Setting env variable: "+str(key)+"="+str(handler_spec['env'][key]))
                            os.environ[key] = handler_spec['env'][key]

                # Handle SLURM_KILL_BAD_EXIT default
                if 'SLURM_KILL_BAD_EXIT' not in os.environ:
                    logging.info("Setting env variable: SLURM_KILL_BAD_EXIT=1 (was unset; to override, set to 0 before invoking mpprun)")
                    os.environ['SLURM_KILL_BAD_EXIT'] = '1'

                # Handle threading
                if args.cpus_per_task is None:
                    threads = 1
                else:
                    if not args.no_smt:
                        threads = args.cpus_per_task
                    else:
                        threads = max(1,args.cpus_per_task//hwinfo.smt_threads_per_physical_core)

                # Always make sure OMP_NUM_THREAD is set (but if set by the user, leave it alone)
                set_thread_vars = []
                unset_thread_vars = []
                if "OMP_NUM_THREADS" not in os.environ or ("OMP_NUM_THREADS" in os.environ and os.environ["OMP_NUM_THREADS"] == ""):
                    os.environ["OMP_NUM_THREADS"] = str(threads)
                else:
                    set_thread_vars += ["OMP_NUM_THREADS"]

                # If OMP_NUM_THREADS, or any of the other threading variables are already set by the user, leave all of them alone,
                # otherwise, set them all to the same value as OMP_NUM_THREADS.
                other_thread_vars = ["OPENBLAS_NUM_THREADS", "MKL_NUM_THREADS", "NUMEXPR_NUM_THREADS"]
                for thread_var in other_thread_vars:
                    if thread_var in os.environ.keys():
                        set_thread_vars += [thread_var]
                    else:
                        unset_thread_vars += [thread_var]

                if len(set_thread_vars) == 0:
                    # No threading variables set at all
                    for thread_var in other_thread_vars:
                        os.environ[thread_var] = str(threads)
                    logging.info("none of the known threading env variables set, all set by mpprun: " + ", ".join([x + "="+str(os.environ[x]) for x in unset_thread_vars + ["OMP_NUM_THREADS"]]))
                elif "OMP_NUM_THREADS" in set_thread_vars:
                    # OMP_NUM_THREADS was set
                    if len(unset_thread_vars) > 0:
                        logging.info("threading env variables set by user: " + ", ".join([x + "="+str(os.environ[x]) for x in set_thread_vars]) + "; left unset by mpprun: "+ ", ".join(unset_thread_vars))
                    else:
                        logging.info("all known threading env variables set by user: " + ", ".join([x + "="+str(os.environ[x]) for x in set_thread_vars]))
                else:
                    # OMP_NUM_THREADS was not set
                    if len(unset_thread_vars) > 0:
                        logging.info("threading env variables set by user: " + ", ".join([x + "="+str(os.environ[x]) for x in set_thread_vars]) + "; set by mpprun: OMP_NUM_THREADS="+os.environ["OMP_NUM_THREADS"] + "; left unset by mpprun: "+ ", ".join(unset_thread_vars))
                    else:
                        logging.info("threading env variables set by user: " + ", ".join([x + "="+str(os.environ[x]) for x in set_thread_vars]) + "; set by mpprun: OMP_NUM_THREADS="+os.environ["OMP_NUM_THREADS"])

                if 'launch_args' in handler_spec:
                    launch_args = handler_spec['launch_args']
                else:
                    launch_args = []

                launcher_exec_path = shutil.which(handler_spec["launcher"])
                if launcher_exec_path is None:
                    raise Exception("Launcher '"+handler_spec["launcher"]+"' not found")

                executable_exec_list = [y for x in args.extra_launch_args for y in shlex.split(x)] + [args.exec_path] + args.arguments
                launcher_exec_list = [launcher_exec_path] + launch_args

                # Handle container wrapper, if requested
                if args.container:
                    container_wrapper = "hpc_container_exec"
                    container_exec_path = shutil.which(container_wrapper)
                    if container_exec_path is None:
                        raise Exception("Container execution wrapper '"+container_wrapper+"' not found")
                    launcher_exec_list += [container_exec_path, args.container]

                exec_list = launcher_exec_list + executable_exec_list

                logging.info("mpprun "+__version__+" executing: "+(" ".join(exec_list)))

                sysmsg = "MPPRUN: "+" ".join([shlex.quote(x) for x in sys.argv])+" => "
                sysmsg += " ".join([shlex.quote(x) for x in exec_list])

                if 'env' in handler_spec:
                    sysmsg += " (ENV: "+str(handler_spec['env'])+")"
                if 'modules' in handler_spec:
                    sysmsg += " (MODULES: "+str(handler_spec['modules'])+")"

                if 'syslog' in handler_spec:
                    sysmsg += " "+" ".join(handler_spec['syslog'])

                if "SLURM_JOB_ID" in os.environ:
                    jid = os.environ["SLURM_JOB_ID"]
                elif "SLURM_JOBID" in os.environ:
                    jid = os.environ["SLURM_JOBID"]
                else:
                    jid = "[unknown]"
                sysmsg += " [SLURM_JOB_ID: "+str(jid)+"]"

                syslog.syslog(sysmsg)

                if args.debug:
                    logging.debug("Executing in envronment: "+pformat(dict(os.environ)))

                try:
                    os.execv(launcher_exec_path,exec_list)
                except FileNotFoundError as e:
                    raise ExceptionWrapper("File not found error when executing: "+(" ".join(exec_list)), e) from e
                raise Exception("Internal error: exec failed for unknown reason")

            except Exception as e:
                raise ExceptionWrapper("Failed to launch executable", e) from e

        else:

            try:

                print()
                print("Found handlers (in order of priority, force use with --handler argument):")
                print()
                for handler in discovered_handlers:
                    handler_name = '_'.join(handler.split('_')[3:])
                    print("  * "+str(handler_name))
                print()

                print("Viable:")
                print()
                for handler_spec in handlers_specs:
                    if 'viable' in handler_spec and handler_spec['viable']:
                        print(" * "+str(handler_spec['handler'])+" (launcher: "+handler_spec['launcher']+"):")
                        del handler_spec['viable']
                        del handler_spec['handler']
                        del handler_spec['launcher']
                        if args.log_level >= logging.INFO:
                            del handler_spec['syslog']
                        for key in handler_spec:
                            if handler_spec[key]:
                                print('    - '+str(key)+": "+str(handler_spec[key]))
                        print()

                print()
                print("Not viable:")
                print()
                for handler_spec in handlers_specs:
                    if 'viable' in handler_spec and not handler_spec['viable']:
                        print(" * "+str(handler_spec['handler'])+" (launcher: "+handler_spec['launcher']+"):")
                        if 'reason' in handler_spec:
                            print('    - REASON: '+handler_spec['reason'])
                            del handler_spec['reason']
                        del handler_spec['viable']
                        del handler_spec['handler']
                        del handler_spec['launcher']
                        if args.log_level >= logging.INFO:
                            del handler_spec['syslog']
                        for key in handler_spec:
                            if handler_spec[key]:
                                print('    - '+str(key)+": "+str(handler_spec[key]))
                        print()
                print()

            except Exception as e:
                raise ExceptionWrapper("Failed to print info", e) from e

    except Exception as e:
        if args.debug:
            raise
        else:
            print(e)
            return 1

def main():
    try:
        args = parse_args(sys.argv)
    except Exception as e:
        print("mpprun internal error when parsing command line arguments: " +type(e).__name__+": "+str(e)+'.', file=sys.stderr)
        if "-d" in sys.argv or "--debug" in sys.argv:
            raise
        return 1

    return cli(args)
