#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.

import os, re, logging

class Hwinfo():

    def __init__(self):
        self.read_cpu_info()
        self.read_numa_layout()
        self.find_sysname()

    def read_cpu_info(self):

        self.cpu_count = 0
        self.cpus = []
        self.physical_cpu_count = 0
        self.physical_cpus = []

        # Get simplified topology info
        if os.path.exists('/sys/devices/system/cpu/cpu0/topology/thread_siblings_list'):
            with open('/sys/devices/system/cpu/cpu0/topology/thread_siblings_list') as f:
                self.smt_threads_per_physical_core = len(f.read().split(","))
        else:
            self.smt_threads_per_physical_core = None
            logging.warning("Cannot access hardware info about smt layout (missing: /sys/devices/system/cpu/cpu0/topology/thread_siblings_list)")

        # Get full topology info
        for f in os.listdir('/sys/devices/system/cpu/'):
            if re.match('^cpu[0-9]+$', f):
                self.cpus += [f]
                self.cpu_count += 1
                if os.path.exists('/sys/devices/system/cpu/'+f+'/topology/thread_siblings_list'):
                    with open('/sys/devices/system/cpu/'+f+'/topology/thread_siblings_list') as f2:
                        siblings = f2.read().strip().split(",")
                        # If not all cpus have the same number of smt threads, warn and nuke the simplified topoplgy info
                        if self.smt_threads_per_physical_core != None and len(siblings) != self.smt_threads_per_physical_core:
                            logging.warning("Machine reports non-consistent smt layout (the number of threads varies per cpu)")
                            self.smt_threads_per_physical_core = None
                        # Take the lowest cpu id for SMT siblings as a physical cpu id
                        physical_cpu = sorted(siblings)[0]
                        if physical_cpu not in self.physical_cpus:                            
                            self.physical_cpus += [physical_cpu]
                            self.physical_cpu_count += 1
                else:
                    self.physical_cpus = None
                    self.physical_cpu_count = None
                    logging.warning("Cannot access hardware info about smt layout (missing: /sys/devices/system/cpu/"+f+"/topology/thread_siblings_list)")
                    break

        if self.cpus is not None:
            self.cpus.sort()
        if self.physical_cpus is not None:
            self.physical_cpus.sort()

    def read_numa_layout(self):

        self.numa_layout = {}
        for f in os.listdir('/sys/devices/system/node/'):
            if re.match('^node[0-9]+$', f):
                self.numa_layout[int(f[4:])] = {}

        for numanode in self.numa_layout:
            self.numa_layout[numanode]['cpus'] = []
            for f in os.listdir('/sys/devices/system/node/node'+str(numanode)):
                if re.match('^cpu[0-9]+$', f):
                    self.numa_layout[numanode]['cpus'] += [int(f[3:])]
            self.numa_layout[numanode]['cpus'].sort()

        self.numa_numanodes_cpu_counts = [len(y['cpus']) for x,y in self.numa_layout.items()]
        self.numa_numanodes_count = len(self.numa_numanodes_cpu_counts)    

        if all(x == self.numa_numanodes_cpu_counts[0] for x in self.numa_numanodes_cpu_counts):
            self.numa_cpus_per_numanode = self.numa_numanodes_cpu_counts[0]
        else:
            self.numa_cpus_per_numanode = None

    def find_sysname(self):
        
        if 'HPC_SSE_SYSNAME' in os.environ:
            self.sysname = os.environ['HPC_SSE_SYSNAME']
        else:
            self.sysname = "unknown"
    
