#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.

"""
A handler that reads a json-formatted .hpcinfo info file that sets the launcher and launcher parameters.

Possible keys to set (all OPTIONAL):

    * launcher (str): the executable name of the underlying (MPI) launcher that this 
         handler uses. The user can via the mpprun argument '--launcher' request a specific
         launcher by name (otherwise the first launcher configuration marked as viable will 
         be used). If the environment described by this handler is used, this is the first 
         part of the executed command.

    * launcher_path (str): if provided, specifies the full path to use to the

    * launch_args (list of str): a list of arguments that are passed to the launcher
         before the executable.

    * env (dict): a dictionary of environment variables to set to the respective values.
         To unset an evironment variable, prefix the key with !, e.g.,
         {"!ONEAPI_DEVICE_SELECTOR": ""} would unset "ONEAPI_DEVICE_SELECTOR".
         To set an environment variable *only* if it is not already set,
         prefix the key with ~, e.g., {"~OMPI_PMI": "pmi2"}.

    * mpprun_default_args (dict): a dictionary modifying default values for arguments to
         mpprun itself. Set the key to the long version of the argument name (e.g., 'ntasks'), 
         and the value to be what you want to set it to. If the user has provided an argument 
         it overrides your setting. You can only set job-affecting arguments, 
         e.g., ntasks, but not, e.g., verbosity.

    * modules (list of str): a list of Lmod modules to load.

    * warnings (list of str): a list of warnings to print out if this handler is used.

    * info (list of str): a list of information to print out if this handler is used.

    * native (bool): indicate whether the specified launcher is the 'native' launcher
      used to build the software.

    * sysname (str): only consider this configuration viable if the sysname of the system
         matches this string.

    * container (str): only consider this configuration viable when launching an exectable
         via the compatibility layer for the system with this sysname.

"""      

import os, logging, re, subprocess, json, copy

def resolve_libmpi(executable):
    found_hpcinfos = []
    
    logging.debug("mpprun srun_hpcinfo handler: check rpath status")
    p = subprocess.run(['objdump', "-x", executable], capture_output=True, text=True)

    if p.returncode != 0:
        raise Exception('objdump error, executable is probably not a binary: '+str(p.stderr.strip()))

    for line in p.stdout.splitlines():
        if line == "":
            continue
        if line.split()[0] in ("RPATH", "RUNPATH"):
            break
    else:
        raise Exception("binary executable does not have rpath set")

    libraries = []
    p = subprocess.run(['ldd', executable], capture_output=True, text=True)
    logging.debug("srun_hpcinfo handler ldd info: "+str(libraries))

    if p.returncode != 0:
        raise Exception("Error when running ldd on "+str(executable))

    for line in p.stdout.splitlines():
        match = re.match(r'\t(.*) => (.*) \(0x', line)
        if match:
            #libraries[match.group(1)] = match.group(2)
            m = match.group(2)
            if m not in libraries:
                libraries += [m]

    found_mpilib = None
    also_check = []
    libraries.sort()

    while libraries:
        lib = libraries.pop(0)
        base = os.path.basename(lib)
        if base.startswith('libmpi'):
            found_mpilib = lib
            if lib.endswith(".so"):
                hpcinfo = lib + ".hpcinfo"
            else:
                prefix, sep, ver = lib.rpartition(".so.")
                hpcinfo = prefix+".so.hpcinfo"

            dirname = os.path.dirname(lib)
            alsocheck = os.path.join(dirname,"libmpi.so")
            if alsocheck != lib and alsocheck not in libraries:
                libraries.insert(0,alsocheck)

            if os.path.exists(hpcinfo):
                if hpcinfo not in found_hpcinfos:
                    found_hpcinfos += [hpcinfo]

    if found_mpilib is None:
        raise Exception("executable not linked to any libmpi*")

    return found_hpcinfos

    
def handle(executable, arguments, args, hwinfo):

    logging.debug("hpcinfo handler invoked on: "+str(executable))
    
    output = {"viable": False, "warnings": [], "info": [], "syslog": [], "native": False, "reason": None, "env": {}, "modules":[], "launch_args": [], "mpprun_default_args":{}}

    # Hardcode the following modules for unloading to avoid unintended
    # conflicts (TODO: improve this handling)
    output["modules"] = [
        "!buildenv-intel",
        "!buildenv-gcc",
        "!OpenMPI",
        "!openmpi",
        "!impi",
        "!foss",
        "!intel",
    ]

    found_hpcinfos = []
    
    # Override the hpcinfo file if there is an application-specific one, and in that case bypass checking libraries
    if os.path.exists(executable + ".hpcinfo"):
        found_hpcinfos += [executable + ".hpcinfo"]

    # Also allow for one at the final chain of symlinks
    if os.path.exists(os.path.realpath(executable) + ".hpcinfo"):
        found_hpcinfos += [os.path.realpath(executable) + ".hpcinfo"]

    # Get hpcinfo from path of libmpi
    try:
        found_libmpi_hpcinfos = resolve_libmpi(executable)
    except Exception as e:
        reason = str(e)
        # If no other hpcinfos were found, give up
        if len(found_hpcinfos) == 0:
            output['reason'] = reason
            return output
    
    found_hpcinfos += found_libmpi_hpcinfos

    # Read all found hpcinfos
    outputs = []        
    if len(found_hpcinfos) > 0:
        for found_hpcinfo in found_hpcinfos:
            try:
                with open(found_hpcinfo) as f:
                    data_list = json.load(f)
            except json.decoder.JSONDecodeError as e:
                new_output = copy.deepcopy(output)
                new_output['reason'] = "hpcinfo file "+str(found_hpcinfo)+" could not be parsed as json: "+str(e)
                outputs += [new_output]
                continue

            if not isinstance(data_list, list):
                data_list = [data_list]

            for data in data_list:
                new_output = copy.deepcopy(output)
                for key, value in data.items():
                    if key in output and isinstance(output[key], list):
                        new_output[key] += value
                    elif key in output and isinstance(output[key], dict):
                        new_output[key].update(value)
                    else:
                        new_output[key] = value

                new_output['syslog'] += ["(hpc_info handler: read "+str(found_hpcinfo)+")"]
                if 'launcher' not in new_output:
                    new_output['reason'] = "hpcinfo file(s) found and read, but it did not have a launcher specified!: "+str(found_hpcinfos)
                else:
                    new_output['viable'] = True
                outputs += [new_output]
                    
    else:
        output['reason'] = "no hpcinfo file found for "+str(executable)+" or its mpi library"
        return output
    
    return outputs
