#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.

"""
This is a template mpprun handler.

Copy it into a name mpprun_hander_<order>_<handler namer>.py where:

  <order>: integer, lower means this handler has higher priority than other
           handlers with higher values.

  <handler name>: string, should be composed by the launcher name + "_" + some
           indication of on what basis the handler deduces the runtime
           environment to launch the job in.

An mpprun handler defines one function 'handle' taking three arguments:

  - executable (str): full path to the executable (binary or script) being executed by mpprun
  - arguments (list of str): the list of arguments that will be passed to that executable
  - args (argparse object): the parsed arguments passed to mpprun

The handler must return:

  - output (dict): a dictionary that describes what launcher, with what arguments,
      in what environment the executable should be launched. It has the following format:

    REQUIRED KEYS:

    * launcher (str): the underlying (MPI) launcher that this handler uses.
         If the environment described by this handler is used, this
         is the first part of the executed command.

    * viable (boolean): indicates if the executable can be launched using
         the environment described by this handler. (False this means mpprun
         will move on to the next handler.)

    OPTIONAL KEYS:

    * warnings (list of str): a list of warnings to print out if this handler is used.

    * info (list of str): a list of information to print out if this handler is used.

    * reason (str): a string explaining why this handler has returned 'viable': False.

    * launch_args (list of str): a list of arguments that are passed to the launcher
         before the executable.

    * mpprun_default_args (dict): a dictionary defining default values for mpprun arguments.
         Set the key to the argparse long version of the argument name, and the value
         to be what you want to set it to. If the user provides the argument it overrides
         your setting. Note: you can only set job-affecting arguments, e.g., ntasks, 
         but not, e.g., verbosity.

    * env (dict): a dictionary of environment variables to set to the respective values.
         To unset an evironment variable, prefix the key with !, e.g.,
         {"!ONEAPI_DEVICE_SELECTOR": ""} would unset "ONEAPI_DEVICE_SELECTOR".

    * modules (list of str): a list of Lmod modules to load.

    * native (bool): indicate whether the specified launcher is the 'native' launcher
      used to build the software.

Note:

- The handler should process the args.ntasks argument into the apropriate way to set a specific 
  amount MPI tasks via launch_args or env.

- The handler is allowed to instead return a list of dictionaries on the above format, in case
  it can derive multiple ways to launch the software.

- The handler is supposed to use the standard python 'logging' module to produce informative output
  with 'logging.debug("...") for things shown only at increased verbosity, and
  'logging.info("...")' for things shown per default. Guideline: always start the handler
  with a single logging.debug("") to describe that the handler is invoked.
"""

import logging

def handle(executable, arguments, args):

    logging.debug("template handler: invoked on: "+str(executable))

    output = {"launcher": "dummy", "viable": False, "warnings":[], "info": []}

    output['reason'] = "This is just a template handler, it is never viable."

    return output
