#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version
#
# This script updates and makes available the latest EB version

# Stop in case of error
set -e

echo "Make sure the dev/... repositories all are checked out to the version you want to install."
echo "Ready? [y/n]"
read READY
if [ "$READY" != "y" ]; then
    echo "Not ready, exiting."
    exit 1
fi

HERE="$(pwd -P)"

print_devel_module()
{
cat <<EOF
#%Module

proc ModulesHelp { } {
    puts stderr {   EasyBuild is a software build and installation framework
written in Python that allows you to install software in a structured,
repeatable and robust way. - Homepage: https://easybuilders.github.io/easybuild/
}
}

module-whatis {EasyBuild is a software build and installation framework
written in Python that allows you to install software in a structured,
repeatable and robust way. - Homepage: https://easybuilders.github.io/easybuild/

This module provides the development version of EasyBuild.
}


EOF
}

GITVER=$(cd "dev/easybuild-framework"; git log -1 | grep -m 1 "release EasyBuild" | sed -n 's/ *release EasyBuild v\?\([0-9.]*\) */\1/p')
if [ -z "$GITVER" ]; then
    echo "Could not determine EasyBuild release number from comment text of last commit."
    echo "(Either the repos are not in a correct state, or the logic of this script needs to be updated)"
    exit 1
fi
GITHASH=$(cd "dev/easybuild-framework"; git log --pretty=format:'%h' -n 1)
if [ -z "$GITHASH" ]; then
    echo "Could not derive a hash number for the git develop repo state."
    exit 1
fi

INSTALL_DIR="$HERE/${GITVER}-hpc${GITHASH}"
EB_MODULE_NAME="EasyBuild/${GITVER}-hpc${GITHASH}"

if [ -e "${INSTALL_DIR}" ]; then
    echo "This version is already installed."
    exit 0
fi

echo "Copying $HERE/dev to $HERE/${GITVER}-hpc${GITHASH}, this may take a while..."
cp -rp "dev" "${GITVER}-hpc${GITHASH}"
echo "Removing unnecessary things inside ${GITVER}-hpc${GITHASH}/, this may take a while..."
rm -rf "./${GITVER}-hpc${GITHASH}/"{vsc-base,vsc-install,easybuild-framework,easybuild-easyblocks,easybuild-easyconfigs}"/.git"

mkdir -p /software/sse2/generic/modules/buildtool-easybuild

EB_VERSION="${GITVER}-hpc${GITHASH}"

cat - > "/software/sse2/generic/modules/buildtool-easybuild/.${EB_VERSION}.lua" <<EOF
help([==[

Description
===========
EasyBuild is a software build and installation framework
written in Python that allows you to install software in a structured,
repeatable and robust way.

More information
================
 - Homepage: https://easybuilders.github.io/easybuild/
]==])

whatis([==[Description: EasyBuild is a software build and installation framework.]==])
whatis([==[Homepage: https://easybuilders.github.io/easybuild/]==])

conflict("buildtool-easybuild")

local home    = os.getenv("HOME")
local ebmods  = pathJoin(home,"EasyBuild","modules","all","Core")
local logname  = os.getenv("LOGNAME")

setenv("EASYBUILD_PREFIX", pathJoin(home, "EasyBuild"))
setenv("HPC_EASYBUILD_PREFIX","/software/sse2/" .. os.getenv("HPC_SSE_SYSNAME") .. "/easybuild/pure")
setenv("EASYBUILD_MODULES_TOOL","Lmod")
setenv("EASYBUILD_RPATH","1")
setenv("EASYBUILD_FILTER_ENV_VARS","LD_LIBRARY_PATH")
setenv("EASYBUILD_FIXED_INSTALLDIR_NAMING_SCHEME","1")
setenv("EASYBUILD_HIDE_DEPS","impi,imkl,ifort,icc,iccifort")
setenv("EASYBUILD_HIDE_TOOLCHAINS","icc,ifort,iimpi,iccifort")
setenv("EASYBUILD_MODULE_NAMING_SCHEME","HierarchicalMNS")
setenv("EASYBUILD_FIXED_INSTALLDIR_NAMING_SCHEME","1")
setenv("EASYBUILD_BUILDPATH",pathJoin("/tmp",logname))
setenv("EASYBUILD_RECURSIVE_MODULE_UNLOAD","1")
setenv("EASYBUILD_ROBOT_PATHS",pathJoin("/software/sse2/",os.getenv("HPC_SSE_SYSNAME"),"/easybuild/pure/ebfiles_repo") .. ":/software/sse2/generic/manual/easybuild/$EB_VERSION/easybuild-easyconfigs/easybuild/easyconfigs")
-- Ensure that software and module folders are group writable
setenv("EASYBUILD_SET_GID_BIT","1")
setenv("EASYBUILD_GROUP_WRITABLE_INSTALLDIR","1")
setenv("EASYBUILD_UMASK","002")

prepend_path("MODULEPATH", "/software/sse2/" .. os.getenv("HPC_SSE_SYSNAME") .. "/easybuild/pure/modules/all/Core")
prepend_path("MODULEPATH", ebmods)

local root = "${INSTALL_DIR}"

prepend_path("PATH",pathJoin(root,"easybuild-framework"))
prepend_path("PYTHONPATH",pathJoin(root,"vsc-base/lib"))
prepend_path("PYTHONPATH",pathJoin(root,"vsc-install/lib"))
prepend_path("PYTHONPATH",pathJoin(root,"easybuild-framework"))
prepend_path("PYTHONPATH",pathJoin(root,"easybuild-easyblocks"))
prepend_path("PYTHONPATH",pathJoin(root,"easybuild-easyconfigs"))

EOF

echo
echo "=== Run 'module load buildtool-easybuild/.${EB_VERSION}' to try the new version of EasyBuild."
echo "=== If you are happy with the install, unhide that module"
echo

exit 0
