#!/usr/bin/env python3
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   weizhang@nsc: 2023-10-27

"""
ModuleNameGenerator - A simple GUI tool for generating module names, adjust the `field_names` list as needed.

Functionality:
- Allows users to input various fields related to a software module.
- Generates a standardized module name based on the input fields.

Usage:
1. Run the script.
2. Fill in the fields as prompted.
3. Click the 'Generate' button to get the standardized module name.

Author: Wei Zhang
Date: 2023-10-27
"""

import sys

try:
    from PyQt5.QtWidgets import (QApplication, QWidget, QVBoxLayout, QHBoxLayout, QLineEdit, QPushButton, 
                             QTextEdit, QLabel, QMessageBox)
    from PyQt5.QtGui import QFont
    from PyQt5.QtCore import Qt
except ModuleNotFoundError:
    print("You need access to a Python with PyQt5 modules. For example, do:")
    print("")
    print("  module load Python/.3.10.4-env-nsc1-gcc-2022a-eb-el7")
    print("")
    print("then re-execute this tool.")
    exit(1)

class ModuleNameGenerator(QWidget):
    def __init__(self):
        super().__init__()

        # Initialize UI
        self.init_ui()

    def init_ui(self):
        # Set window properties
        self.setWindowTitle('Module Name Generator')
        self.resize(600, 400)  # Set initial window size

        # Set global font size
        font = QFont()
        font.setPointSize(14)
        self.setFont(font)

        # Create main layout
        main_layout = QVBoxLayout()

        # Field name list
        self.field_names = ["Software", "Version", "Feature Tag", "hpc1, hpc2, ...", "Toolchain", "Tag"]
        
        # Initial values list
        initial_values = ["Siesta", "4.1.5", "b4", "hpc1", "intel-2018a-eb", "compat-el7"]

        # Adjust initial_values based on the length of field_names
        while len(initial_values) < len(self.field_names):
            initial_values.append('')

        # Create input boxes and add them to the layout
        self.entries = []
        for field_name, initial_value in zip(self.field_names, initial_values):
            # Create horizontal layout with left alignment
            h_layout = QHBoxLayout()
            h_layout.setAlignment(Qt.AlignLeft)
            
            # Create and add label with left alignment
            if field_name in ["Software", "Version"]:
                label = QLabel(f"{field_name}<span style='color: red'>*</span>")
            else:
                label = QLabel(field_name)
            label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
            h_layout.addWidget(label, 1)

            # Create and add input box
            entry = QLineEdit(self)
            entry.setFont(font)  # Set input box font size
            entry.setText(initial_value)  # Set initial value
            self.entries.append(entry)
            h_layout.addWidget(entry, 3)

            # Add horizontal layout to main layout
            main_layout.addLayout(h_layout)

        # Create button and connect to concatenate function
        self.concat_button = QPushButton('Generate', self)
        self.concat_button.setFont(font)  # Set button font size
        self.concat_button.clicked.connect(self.concatenate_strings)
        main_layout.addWidget(self.concat_button)

        # Create Help button
        self.help_button = QPushButton('Help', self)
        self.help_button.setFont(font)
        self.help_button.clicked.connect(self.show_help)
        main_layout.addWidget(self.help_button)

        # Create text box to display result
        self.result_text = QTextEdit(self)
        self.result_text.setFont(font)  # Set text box font size
        self.result_text.setReadOnly(True)
        main_layout.addWidget(QLabel("Module name:"))
        main_layout.addWidget(self.result_text)

        # Add a note about required fields
        required_note = QLabel("<span style='color: red'>*</span> Fields with a red asterisk are required.")
        required_note.setFont(QFont("Arial", 10))
        main_layout.addWidget(required_note)

        # Set main layout
        self.setLayout(main_layout)

    def concatenate_strings(self):
        # Reset all entry styles to clear previous highlights
        for entry in self.entries:
            entry.setStyleSheet("")

        # Get values of each field and strip spaces
        values = [entry.text().strip() for entry in self.entries]

        # Check for empty or invalid fields
        if not values[0] or ' ' in values[0].strip():
            self.show_error("Software field cannot be empty and cannot contain spaces in between!", self.field_names[0])
            return
        if not values[1] or ' ' in values[1].strip():
            self.show_error("Version field cannot be empty and cannot contain spaces in between!", self.field_names[1])
            return
        if values[4] and not values[4].startswith(('intel', 'gcc', 'nvhpc', 'amd')):
            self.show_error("Toolchain must start with 'intel', 'gcc', 'nvhpc', or 'amd' or be empty!", self.field_names[4])
            return
        for index, value in enumerate(values):  # Check all fields
            if ' ' in value.strip():
                self.show_error(f"{self.field_names[index]} field cannot contain spaces in between!", self.field_names[index])
                return

        # Use / to connect the first two fields
        before_slash = '/'.join(values[:2])

        # Use - to connect the rest, but filter out empty values
        after_slash = '-'.join(filter(None, values[2:]))

        # Combine the two parts
        result = before_slash
        if before_slash and after_slash:
            result += '-'
        result += after_slash

        # Display result
        self.result_text.setPlainText(result)

    def show_help(self):
        explanations = {
            "Software": "Name of the software. (Required)",
            "Version": "Version of the software. (Required)",
            "Feature Tag": "Specific feature tag for the software. Can contain spaces if empty. (Optional)",
            "hpc1, hpc2, ...": "Additional tags. Can contain spaces if empty. (Optional)",
            "Toolchain": "Must start with 'intel', 'gcc', 'nvhpc', or 'amd' or be empty. Can contain spaces if empty. (Optional)",
            "Tag": "Additional tag. (Optional)"
        }
        
        help_text = "\n\n".join([f"{field}: {explanations.get(field, 'Description not available.')}" for field in self.field_names])        
        # Create a message box with larger font and size for help
        msgBox = QMessageBox(self)
        msgBox.setWindowTitle("Help")
        msgBox.setText(help_text)
        msgBox.setFont(QFont("Arial", 14))
        msgBox.resize(600, 350)
        msgBox.exec_()

    def show_error(self, message, field_name):
        # Find the erroneous entry and set its border to red
        index = self.field_names.index(field_name)
        self.entries[index].setStyleSheet("border: 2px solid red;")

        # Create a message box with larger font and size for errors
        errorBox = QMessageBox(self)
        errorBox.setWindowTitle(f"Error in {field_name}")
        errorBox.setText(message)
        errorBox.setIcon(QMessageBox.Warning)
        errorBox.setFont(QFont("Arial", 16))
        errorBox.resize(600, 200)
        errorBox.exec_()

if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = ModuleNameGenerator()
    window.show()
    sys.exit(app.exec_())

