#!/usr/bin/python3 -E
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2022 initial version

from __future__ import print_function

import os, sys, re, argparse, glob
import yaml, json, pprint

sysname = os.environ["HPC_SSE_SYSNAME"]

easyconfigs_path = '/software/sse2/'+sysname+'/easybuild/pure/easyconfigs'

expected_software_keys = set(['software_name','software_desc_full','software_desc','homepage','classes','support','nsc_homepage'])
expected_module_keys = set(['name','module_desc','deprecated','software_version','build_features','doc_path','doc_url','compatibility'])

# Get this string from 'eb --show-default-moduleclasses' and put our categories manually in third column
EB_MODULE_CLASSES_STRING="""
base:      Default module class                                            :misc
bio:       Bioinformatics, biology and biomedical                          :bio
cae:       Computer Aided Engineering (incl. CFD)                          :cae
chem:      Chemistry, Computational Chemistry and Quantum Chemistry        :chem
compiler:  Compilers                                                       :devel
data:      Data management & processing tools                              :data
debugger:  Debuggers                                                       :devel
devel:     Development tools                                               :devel
geo:       Earth Sciences                                                  :geo
ide:       Integrated Development Environments (e.g. editors)              :devel
lang:      Languages and programming aids                                  :devel
lib:       General purpose libraries                                       :devel
math:      High-level mathematical software                                :math
mpi:       MPI stacks                                                      :devel
numlib:    Numerical Libraries                                             :devel
perf:      Performance tools                                               :devel
phys:      Physics and physical systems simulations                        :phys
system:    System utilities (e.g. highly depending on system OS and hardware) :tools
toolchain: EasyBuild toolchains                                            :devel
tools:     General purpose tools                                           :tools
vis:       Visualization, plotting, documentation and typesetting          :vis
"""
eb_module_classes = dict([(y[0].strip(),y[1].strip()) for y in [x.split(":") for x in EB_MODULE_CLASSES_STRING.strip().splitlines()]])
eb_module_map = dict([(y[0].strip(),y[2].strip()) for y in [x.split(":") for x in EB_MODULE_CLASSES_STRING.strip().splitlines()]])

nsc_classes = {
    'bio':{},
    'cae':{},
    'chem':{},
    'data':{},
    'devel':{},
    'geo':{},
    'math':{},
    'phys':{},
    'tools':{},
    'vis':{},
    'misc':{},
}

nsc_classes_names = {
    'bio':'Bioinformatics, biology and biomedical',
    'cae':'Computer Aided Engineering (incl. CFD)',
    'chem':'Chemistry, Computational Chemistry and Quantum Chemistry',
    'data':'Data management and processing tools',
    'devel':'Software Development',
    'geo':'Earth Sciences',
    'math':'Mathematics',
    'phys':'Physics and physical systems simulations',
    'tools':'General purpose tools',
    'vis':'Visualization, plotting, documentation and typesetting',
    'misc':'Uncategorized'
}

nsc_default_values = {
    'support': 'tier3',
    'classes': ["misc"]
}

yaml.SafeDumper.org_represent_str = yaml.SafeDumper.represent_str

def grab_yaml_header(content):
    # Look for yaml header
    content_lines = content.splitlines()
    match = re.match('--( +)--- *$', content_lines[0])
    if match:
        header_spaces = match.group(1)
        prefix = "^--" + header_spaces
        yamlstr = ""
        for line in content_lines[1:]:
            if re.match("^-- *$",line):
                yamlstr += "\n"
                continue
            if re.match(prefix + "(---|\.\.\.)",line):
                break
            if re.match(prefix,line):
                yamlstr += line[2+len(header_spaces):] + "\n"
                continue
            break
        #print("EXTRACTED",yamlstr)
        return yaml.safe_load(yamlstr)
    return None

def grab_data_from_module_file(content):

    description = None
    helpstr = None
    homepage = None
    whatis = []
    whatis_categories = {}
    summary = None

    #description_match = re.compile(r"-- NSC DESCRIPTION BEGIN *\n(.*?)\n-- NSC DESCRIPTION END",re.DOTALL)
    #match = description_match.match(content)
    #if match:
    #  description = match.group(1).strip()
    #  description = " ".join([x[3:] for x in description.split("\n")])

    #homepage_match = re.compile(r"-- NSC HOMEPAGE *(.*)\n")
    #match = homepage_match.match(content)
    #if match:
    #  homepage = match.group(1).strip()

    help_match = re.compile(r"help\( *\[=?=?\[(.*?)\]=?=?\]",re.DOTALL)
    match = help_match.match(content)
    if match:
      helpstr = match.group(1).strip()

    if helpstr is not None and description is None:
      description_match = re.compile(r"Description\n===* *\n(.*?)(\n===*|\Z)",re.DOTALL)
      match = description_match.match(helpstr)
      if match:
        description = match.group(1)
        if len(match.group(2)) > 0:
          description = description[:description.rfind('\n')].strip()
        else:
          description = description.strip()
        description = description.replace("\n"," ").strip()

    if helpstr is not None and homepage is None:
      homepage_match = re.compile(r"- Homepage: *(.*)$")
      match = homepage_match.match(helpstr)
      if match:
        homepage = match.group(1)

    whatis_match = re.compile(r"whatis\( *\[==\[(.*?)\]==\]",re.DOTALL)
    match = whatis_match.findall(content)
    if match:
      whatis = [x.replace("\n"," ") for x in match]

    if len(whatis) == 0:
        whatis_match = re.compile(r"whatis\( *"+'"'+r"\[(.*?)\]"+'"',re.DOTALL)
        match = whatis_match.findall(content)
        if match:
            whatis = [x.replace("\n"," ") for x in match]

    whatis_categories_match = re.compile(r"whatis\( *\[==\[([^ ]*): *(.*?)\]==\]",re.DOTALL)
    matches = whatis_categories_match.findall(content)
    if matches:
        for match in matches:                    
            whatis_categories[match[0].lower()] = match[1].replace("\n"," ")

    if homepage is None and 'homepage' in whatis_categories:
        homepage = whatis_categories['homepage']

    categories = nsc_default_values['classes']
    easyconfigs = glob.glob(os.path.join(easyconfigs_path,"*",module_name,module_name+"-*.eb"))
    if len(easyconfigs) > 0:
      with open(os.path.join(easyconfigs[-1])) as f:
        content = f.read()
      category_match = re.search("moduleclass *= ('([^']*)'|\"(^[^\"]*)\")",content)
      if category_match:
        category = category_match.group(2) if category_match.group(2) != "" else category_match.group(3)
        category = eb_module_map[category]
        categories = [category]

    if description is None:
      if helpstr is not None and helpstr != "":
        description = helpstr

        last_line = description.split()[-1]
        if last_line.startswith("http"):
            if homepage is None:
                homepage = last_line
            description = description[:description.rfind('\n')].strip()
        description = description.replace("\n"," ").strip()

      elif whatis is not None and len(whatis) > 0:
        description = "\n\n".join(whatis)
      else:
        print("WARNING",module_name,": module had no description, help or whatis.")
        description = ""

    if 'description' in whatis_categories:
        summary = whatis_categories['description']
    else:
        summary = description

    if helpstr is None:
        print("NOTICE",module_name,": missing help text")

    if len(whatis) == 0:
        print("NOTICE",module_name,": missing whatis text")

    data = {'software_name':module_name, 
            'software_desc_full': description,
            'software_desc': summary,
            'homepage': homepage,
            #'help': helpstr,
            #'whatis': whatis,
            'classes': categories,
            #'class_name': nsc_classes_names[category]
    }

    return data


def repr_str(dumper, data):
    if '\n' in data:
        return dumper.represent_scalar(u'tag:yaml.org,2002:str', data, style='|')
    return dumper.org_represent_str(data)

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Compose json or yaml files based on NSC module tree")
    parser.add_argument('outfile', help='The output file (or with -s, directory)')
    parser.add_argument('-m','--modulespath', help='The directory to get modules from (default: /software/sse2/'+sysname+'/modules)', default="/software/sse2/"+sysname+"/modules")
    parser.add_argument('-s','--separate', help='Generate separate output files for each module', action="store_true", default=False)
    parser.add_argument('-f','--format', help='Output format', default="json", choices=["json", "yaml", "header"])
    args = parser.parse_args()

    if args.outfile is None or args.outfile == "":
      parser.print_help(sys.stderr)
      sys.exit(1)

    if args.separate and not os.path.exists(args.outfile):
      raise Exception("Separate file output selected, and output dir "+args.outfile+" does not exist.")

    if (not args.separate) and args.format == "header":
      raise Exception("Header format requires separate files.")

    sse_modules_path = args.modulespath
    #header_formatted = args.header
    module_names = sorted(os.listdir(sse_modules_path), key=lambda s: s.lower())

    all_data = []

    def version_split(s):
        vals = []
        parts = s.split('.')
        for part in parts:
          try:
            val = "{0:0=10d}".format(int(part))
          except Exception:
            val = part
          vals += [val]
        return vals

    for module_name in module_names:
        print("\n====",module_name)
        software_data = None
      
        moddir = os.listdir(os.path.join(sse_modules_path, module_name))
        
        if 'recommendation.lua' in moddir:
            with open(os.path.join(sse_modules_path, module_name,"recommendation.lua")) as f:
              content = f.read()
            software_data = grab_yaml_header(content)

        all_versions = sorted( [x[:x.rfind(".lua")] for x in moddir if (x.endswith(".lua") and not x == 'recommendation.lua' and not x.startswith("."))] , key=version_split, reverse=True)

        if len(all_versions) == 0 and software_data is None:
          continue

        default_ver = None
        if 'default' in moddir:
          default_ver = os.readlink(os.path.join(sse_modules_path, module_name,"default"))
          if default_ver == 'recommendation.lua':
            with open(os.path.join(sse_modules_path, module_name,"recommendation.lua")) as f:
              content = f.read()
              out = re.search("nsc_message_recommend\(\"(.*)\"\)",content)
              if not out:
                  out = re.search("nsc_message_recommend\('(.*)'\)",content)
              if out:
                default_ver = out.group(1)
                if default_ver == "":
                  default_ver = None
                elif not os.path.exists(os.path.join(sse_modules_path, module_name,default_ver + ".lua")):
                  print("WARNING",module_name,": module contains a recommendation that does not exist:",default_ver)
                  default_ver = None
              else:
                default_ver = None
          else:
            default_ver = default_ver[:default_ver.rfind(".lua")]

        if default_ver == "":
          print("ERROR",module_name,": BROKEN - no default_ver out of:", all_versions, default_ver)
          continue

        module_versions = []
        for version in all_versions:
            print("== reading version:",version)
            with open(os.path.join(sse_modules_path, module_name, version + ".lua")) as f:
                version_content = f.read()
            version_data = grab_yaml_header(version_content)

            if version_data is None:

                #Not really needed, as nothing is used for module_versions
                #version_data = grab_data_from_module_file(version_content):

                module_versions += [
                    {'name':version,
                     'module_desc':None,
                     'deprecated':None,
                     'software_version':None,
                     'build_features':None,
                     'doc_path':None,
                     'doc_url':None,
                 }]                
            else:
                version_data['name'] = version
                keys = set(version_data.keys())
                extra_keys = keys - expected_module_keys
                missing_keys = expected_module_keys - keys
                if len(extra_keys) > 0:
                    print("NOTICE",module_name,version,": Header contains unexpected extra keys",extra_keys)
                if len(extra_keys) > 0:
                    print("WARNING",module_name,version,": Header is missing expected keys",missing_keys)                    
                module_versions += [version_data]

        if software_data is None:

            if len(all_versions)>0:

                if default_ver is None:
                    read_ver = all_versions[0]
                else:
                    read_ver = default_ver

                with open(os.path.join(sse_modules_path, module_name,read_ver + ".lua")) as f:
                    content = f.read()

                data = grab_yaml_header(content)

                if data is None:
                    print("NOTICE",module_name,": No relevant headers found, grabbing data from module contents.")
                    data = grab_data_from_module_file(content)
                else:
                    print("NOTICE",module_name,": Data extracted from header of default version.")

                if data is not None:
                    # Only keep expected keys extracted this way
                    data = { k: data[k] for k in expected_software_keys if k in data }
            else:
                print("NOTICE",module_name,": No software information found anywhere.")
                data = {}

        else:
            print("NOTICE",module_name,": Data extracted from recommended.lua.")
            data = software_data

        if 'software_name' not in data:
            data['software_name'] = module_name

        keys = set(data.keys())
        extra_keys = keys - expected_software_keys
        missing_keys = expected_software_keys - keys
        if len(extra_keys) > 0:
            print("NOTICE",module_name,": header contains unexpected extra keys:",", ".join(extra_keys))
        if len(missing_keys) > 0:
            print("WARNING",module_name,": header is missing keys:",", ".join(missing_keys))
            for add_key in missing_keys:
                if add_key in nsc_default_values:
                    data[add_key] = nsc_default_values[add_key]
                else:
                    data[add_key] = None

        data['module_name'] = module_name
        data['module_versions'] = module_versions
        data['default_version'] = default_ver 

        if 'classes' in data:
            for c in data['classes']:
                if c not in nsc_classes:
                    print("WARNING",module_name,": classes lists unknown class:",c)

        if args.separate:

            if args.format == "yaml" or args.format == "header":

                yaml.add_representer(str, repr_str, Dumper=yaml.SafeDumper)
                with open(os.path.join(args.outfile,module_name+".yaml"),"w") as f:
                    if args.format == "header":
                        del data['module_versions']
                        del data['default_version']                        
                        output = yaml.safe_dump(data, allow_unicode=True, default_flow_style=False)
                        f.write("-- ---\n")
                        for line in output.splitlines():
                            f.write("-- "+line+"\n")
                        f.write("-- ---\n")
                    else:
                        yaml.safe_dump(data, f, allow_unicode=True, default_flow_style=False)
 
            if args.format == "json":
               
                with open(os.path.join(args.outfile,module_name+".json"),"w") as f:
                    json.dump(data, f)

        else:
            all_data += [data]


if not args.separate:

    if args.format == "yaml":

        yaml.add_representer(str, repr_str, Dumper=yaml.SafeDumper)
        with open(args.outfile,"w") as f:
            yaml.safe_dump(all_data, f, allow_unicode=True, default_flow_style=False)
                
    if args.format == "json":
               
        with open(args.outfile,"w") as f:
            json.dump(all_data, f, indent=2)

