#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023-10-05: Add support for a link-time-only LD_LIBRARY_PATH by setting
#     HPC_BUILD_LD_LIBRARY_PATH to handle intel loading ld plugins with
#     unresolved symbols that require LD_LIBRARY_PATH.
#
#   rar@nsc: 2023-08-24: Adjustments to fit the new SSE2 framework. Minor cleanup.
#
#   raber@nsc, cbasu@nsc: 2023-06-22: Adding infrastructure for the new tagging system using ELF
#       sections with the content of envvars available at compile time.
#
#   raber@nsc: 2022-04-07: Edits allowing standalone use, removing tagging and permitting
#       the use of ld.gold.
#
#   rar@nsc: 2018-2020: Edits for tetralith setup.
#
#   cbasu@nsc.liu.se: 2015-02-12: oringal version
#
# ld_wrapper part of the HPC SSE tools
#
# Wrapper for the linker to make sure linked binaries are rpath:ed
# and provided with metadata sections providing enough information
# to determine the runtime environment in which to execute the binary.
#
# The wrapped linker adds named metadata sections to the binary using
# environment variables at compile time for content. The content of these
# metadata sections are used by the mpprun launcher ar runtime to launch MPI
# applications correctly
#
##

set -e

# Protect against recursion
if [ "$HPC_LDWRAP_NORECURSE" == "1" ]; then
    echo "The HPC ldwrap script appears to recurse. Please seek help from support."
    exit 1
fi
export HPC_LDWRAP_NORECURSE=1

# We want to find out the path via which this script was invoked, even if it was a symlink
# (so we can remove it from the PATH variable)
SCRIPT="$(realpath -s "$0")"
SCRIPTPATH="$(dirname "$SCRIPT")"

# Store which linker was requested for potential special treatment
LINKER_REQESTED="$(basename $0)"

if [ "$LINKER_REQESTED" == "x86_64-unknown-linux-gnu-ld.lld" ]; then
	    LINKER_REQESTED="ld.lld"
fi

# Strip the $SCRIPTPATH from PATH to get the next `ld` or `ld.gold` available
OLDPATH="${PATH}"
export PATH="${OLDPATH//"${SCRIPTPATH}:"/}"

# Set which linker should be used. Use absolute path.
LDORIG=$(type -p "${LINKER_REQESTED}")

# Backwards compatibility for ld wrapper flags:
if [ -z "$HPC_LD_FLAG" -a -n "$NSC_LD_FLAG" ]; then
    HPC_LD_FLAG="$NSC_LD_FLAG"
fi
if [ -z "$HPC_LD_EXTRA_LIBPATH" -a -n "$NSC_LD_EXTRA_LIBPATH" ]; then
    HPC_LD_EXTRA_LIBPATH="$NSC_LD_EXTRA_LIBPATH"
fi
if [ -z "$HPC_COMPILER_VERBOSE" -a -n "$NSC_COMPILER_VERBOSE" ]; then
    HPC_COMPILER_VERBOSE="$HPC_COMPILER_VERBOSE"
fi

# Check HPC_LD_FLAG for whether to wrap ld or not, and what directories to include/exclude
if [ -z "${HPC_LD_FLAG+x}" ]; then
    HPC_LD_FLAG=1
fi
if [ "$HPC_LD_FLAG" == 0 ]; then
    # Just hand over to the original linker
    exec "${LDORIG}" "$@"
elif [ "$HPC_LD_FLAG" == 1 ]; then
    # Default rpath-ing option. Here we can put other criterias,
    # e.g., strict compliance with module system etc.
    #exclude_lib_paths=("/lib" "/usr" "/home" "/tmp" "/opt" "/nobackup")
    #exclude_lib_paths=("/lib" "/lib64" "/usr" "/home" "/tmp" "/opt" "/proj")
    EXCLUDE_LIBDIRS="/lib:/lib64:/usr:/home:/tmp:/opt:/proj"
elif [ "$HPC_LD_FLAG" == 2 ]; then
    # More aggressive rpath-ing
    #exclude_lib_paths=("/lib" "/lib64" "/usr" "/tmp" "/opt")
    #exclude_lib_paths=("/tmp" "/opt")
    EXCLUDE_LIBDIRS="/lib:/lib64:/usr:/tmp:/opt"
fi

# Process input arguments
#
# If -o is given, figure out the output binary
#
# Collect all added library directories
#
# Remove the --enable-new-dtags flag, since it creates a copy of rpath to runpath
# If runpath exists in the binary it can be controlled by LD_LIBRARY_PATH.
# We want only rpath in the binary and no runpath.
# To use runpath, disable linker warpper with HPC_LD_FLAG=0
#
# TODO: I think the new default behavior is to put rpaths in RUNPATH, so we may want
# to force --disable-new-dtags now?

LIBDIRS="$LIBRARY_PATH"
BINNAME="a.out"
for (( i=${#@}; i >= 0; i-- )); do
    if [ "${!i}" == "--enable-new-dtags" ]; then
        set -- "${@:1:$(( $i - 1 ))}" "${@:$(( $i + 1 ))}"
        continue
    elif [ "${!i}" == "-o" -o "${!i}" == "--output" ]; then
        NEXT_ARG=$(($i + 1))
	BINNAME="${!NEXT_ARG}"
        continue
    elif [ "${!i}" == "-L" -o "${!i}" == "--library-path" ]; then
        NEXT_ARG=$(($i + 1))
        NEW_LIBDIR="${!NEXT_ARG}"
    elif [ "${!i#--library-path=}" != "${!i}" ]; then
        NEW_LIBDIR="${!i#--library-path=}"
    elif [ "${!i#-L.}" != "${!i}" -o "${!i#-L/}" != "${!i}" ]; then
        NEW_LIBDIR="${!i#-L}"
    else
        continue
    fi
    LIBDIRS="$LIBDIRS:$NEW_LIBDIR"
done

# If LD_LIBRARY_PATH is set we probably should rpath those directories too since they
# may be picked up by the linker
if [ -n "$LD_LIBRARY_PATH" ]; then
    LIBDIRS="$LD_LIBRARY_PATH:$LIBDIRS"
fi

# Remove LIBDIRS that start with excluded paths
LIBDIRS="$(tr : '\n' <<< "$LIBDIRS" | grep -v "^\(${EXCLUDE_LIBDIRS//:/\\|}\)" | tr '\n' :)"
LIBDIRS="${LIBDIRS%:}"

# Append HPC extra libpath if set
if [ -n "$HPC_LD_EXTRA_LIBPATH" ]; then
    LIBDIRS="$LIBDIRS:$HPC_LD_EXTRA_LIBPATH"
fi

# Remove duplicates
LIBDIRS="$(tr : '\n' <<< "$LIBDIRS" | awk '!($0=="" || x[$0]++)' | tr '\n' :)"
LIBDIRS="${LIBDIRS%:}"

if [ -n "$LIBDIRS" ]; then
    RPATH="-rpath=${LIBDIRS//:/ -rpath=}"
else
    RPATH=""
fi

# Debug output if requested via variables being set.

if [ -n "$HPC_COMPILER_VERBOSE" -a "$HPC_COMPILER_VERBOSE" != "false" ]; then
    if [ "$HPC_COMPILER_VERBOSE" = "ERR" ]; then
		echo "INFO: HPC LD WRAPPER prefix args: $RPATH" >&2
		echo "INFO: HPC LD WRAPPER full call:" $LDORIG $RPATH "$@" >&2
    elif [ "$HPC_COMPILER_VERBOSE" = "LOG" ]; then
		echo "INFO: HPC LD WRAPPER prefix args: $RPATH" >> ~/hpc_compiler_wrapper.log
		echo "INFO: HPC LD WRAPPER full call:" $LDORIG $RPATH "$@" >> ~/hpc_compiler_wrapper.log
    else
		echo "INFO: HPC LD WRAPPER prefix args: $RPATH"
		echo "INFO: HPC LD WRAPPER full call:" $LDORIG $RPATH "$@"
    fi
fi

# Handle a user-set LD_LIBRARY_PATH plus our HPC_BUILD_LD_LIBRARY_PATH
if [ -n "$LD_LIBRARY_PATH" ]; then
  HPC_BUILD_LD_LIBRARY_PATH="${HPC_BUILD_LD_LIBRARY_PATH}:${LD_LIBRARY_PATH}"
fi

# Do the linking
LD_LIBRARY_PATH="${HPC_BUILD_LD_LIBRARY_PATH}" "${LDORIG}" ${RPATH} "$@"

# Try to write some additional info to the binary, but don't break on failure
if [ -e "${BINNAME}" ]; then
    # Saftey check, don't objcopy if the file is not created now (older than 1 hour).
    AGE="$(($(date +%s) - $(date +%s -r "$BINNAME")))"
    if [ "$AGE" -lt "3600" ]; then
        objcopy \
	    --add-section .hpc_linktime_path=<(printf "%s\n" "PATH=${PATH}") \
	    --add-section .hpc_linktime_fi_provider_path=<(printf "%s\n" "FI_PROVIDER_PATH=${FI_PROVIDER_PATH}") \
	    --add-section .hpc_linktime_modulepath=<(printf "%s\n" "MODULEPATH=${MODULEPATH}") \
	    --add-section .hpc_linktime_loadedmodules=<(printf "%s\n" "LOADEDMODULES=${LOADEDMODULES}") \
	    "${BINNAME}" || true
    fi
fi
