#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# This script is called from hpc Lmod module to:
#
# - display quota warnings and other useful information.
# - create /proj/*/users/$USER

# If we want to unmute login messages (e.g if we want all users to see
# some new information), change this timestamp to $(date +%s)
FORCE_UNMUTE_LIMIT=1538634514 # 2018-10-04

# Check if we're on a login node. There's currently no canonical way
# to determine this, but a hostname that starts with
# $NSC_RESOURCE_NAME should work for all current clusters that use
# lmod.
on_login_node () {
    if [ ! -z "${NSC_RESOURCE_NAME}" ]; then
	if [[ "$(hostname -s)" =~ ^${NSC_RESOURCE_NAME}[0-9]?$ ]]; then
	    return 0
	fi
    fi
    return 1
}

# Check if the user wants to see information on login
user_wants_login_info () {
    mutefile="${HOME}/.nsc-mute-login-${NSC_RESOURCE_NAME}"
    if [ ! -f "$mutefile" ]; then
	return 0
    else
	muted_at=$(stat --format="%Y" $mutefile)
	if [ $muted_at -lt $FORCE_UNMUTE_LIMIT ]; then
	    return 0
	else
	    return 1
	fi
    fi
}

# This will allow us to set a custom message that will be displayed to
# all users who have not opted out of login information.
print_custom_motd () {
    CUSTOM_MOTDS="/software/sse2/init/custom_motds/"
    CUSTOM_MOTDS_CLUSTER="/software/sse2/init/custom_motds/${NSC_RESOURCE_NAME}"

    # Static messages
    files=$(ls ${CUSTOM_MOTDS}/*-static.txt ${CUSTOM_MOTDS_CLUSTER}/*-static.txt 2>/dev/null | sort -n)
    if [ -n "$files" ]; then
	for file in $files; do
	    cat "$file"
	    echo
	done
    fi

    # Expiring messages
    files=$(ls ${CUSTOM_MOTDS}/*-expires-*.txt ${CUSTOM_MOTDS_CLUSTER}/*-expires-*.txt 2>/dev/null | sort -n)
    if [ -z "$files" ]; then
	return
    fi
    for file in $files; do
	expiredate=$(echo "$file" | sed -e 's/^.*-expires-\(.*\).txt$/\1/')
	now_timet=$(date "+%s")
	expiredate_timet=$(date -d "$expiredate" "+%s" 2>/dev/null)
	if [ $? -eq 0 ] && [ ! -z "$expiredate_timet" ]; then
	    if [ "$now_timet" -lt "$expiredate_timet" ]; then
		echo -e "**** Important information (expires $expiredate):"
		cat "$file"
		echo
	    fi
	else
	    logger -t warnings_and_welcome_messages "Bad custom motd filename $file - cannot parse expire time: $expiredate"
	fi
    done
}

# Ring bell once and output prominent warning to get user's attention
warnline () {
	tput bel; tput setaf 1; for a in $(seq 1 8); do echo -n "WARNING! "; done; echo; tput sgr0
}

# Create personal /proj directories and save output (only outside
# job environment, i.e on login node and system server)
if on_login_node ; then
    if [ -e /software/tools/bin/check_and_create_user_projects_dirs ]; then
	personal_proj_dirs=$(/software/tools/bin/check_and_create_user_projects_dirs)
    fi
fi

# Skip the rest of the script if we have no tty to display the output
# on. Prevents problems when running e.g "ssh tetralith 'somecommand' > somecommand.out"
tty -s || exit 0

#
# Optional information section
#
# Skip if in job environment or if the user does not want to see
# this.
#
if on_login_node && user_wants_login_info; then
    echo "Welcome to ${SNIC_SITE^^} and ${NSC_RESOURCE_NAME^}!"
    echo

    if [ ! -z "${personal_proj_dirs}" ]; then
	cat <<EOF
**** Project storage directories available to you:
${personal_proj_dirs}

EOF
    fi

    cat <<EOF
**** Documentation and getting help:
https://www.nsc.liu.se/support/systems/${NSC_RESOURCE_NAME}-getting-started/
https://www.nsc.liu.se/support

EOF
    print_custom_motd

    cat <<EOF
(Run "nsc-mute-login" to not show this information)

EOF
fi

#
# Mandatory information section: quota warnings etc.
# Skip only if in job environment.
#
if on_login_node; then
    if [ -e /software/tools/bin/storagequota ]; then
	# By using a bash coprocess and the timeout (-t) argument to
	# 'read', we make sure the login does not hang for too long if
	# storagequota is hanging.  storagequota will hang if any
	# filesystem it wants to check the quota on is hanging.
	coproc /software/tools/bin/storagequota --warn-only
	read -d '' -r -t 3 -u ${COPROC[0]} quotawarning
	if [ ! -z "${quotawarning}" ]; then
	    warnline
	    echo "${quotawarning}"
	    warnline
	    echo
	fi
    fi
    if [ -e /software/tools/bin/tmpquota ]; then
	quotawarning=$(/software/tools/bin/tmpquota)
	if [ ! -z "${quotawarning}" ]; then
	    warnline
	    echo "${quotawarning}"
	    warnline
	    echo
	fi
    fi
fi
