#!/usr/bin/env python3
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version

import os, sys, re, pathlib, argparse

def module_show(d, *args):
    output = os.popen('module use %s; module show %s' % (d, " ".join(args)),)
    return output

def get_paths(d, module):
    paths = []
    out = module_show(d, module)
    for line in out:
      m = re.search(" *prepend_path\([\"']PATH[\"'],['\"](.+?)['\"]", line)
      if m:
        path = m.group(1)
        paths += [path]

    return paths


def process_module(md, d, dout, dwrap, wrapper, tag, name, dry=False):  

    modulename = os.path.basename(d)

    inname = os.path.join(d,name)
    outdir = os.path.join(dout, modulename)
    verfile_extless = os.path.splitext(name)[0]
    outname = os.path.join(outdir,verfile_extless+tag+".lua")

    print("== Checking",name)
    paths = get_paths(md, modulename+"/"+name)

    with open(inname, "rt") as f:
        if not os.path.exists(outdir):
          if not dry:
            os.mkdir(outdir)
          else:
            print("mkdir",outdir)
        if not pathlib.Path(outname).exists():
          if not dry:
            with open(outname, "wt") as fout:
              for line in f:
                if 'prepend_path("PATH",' in line:
                  line = line.replace('prepend_path("PATH",','prepend_path("PATH", "'+dwrap+'" .. ')
                fout.write(line)
          else:
            print("create",outname)
        else:
          print("Module file already exists:",outname)

    for path in paths:
      relpath = os.path.relpath(path, start="/")
      outwrapdir = os.path.join(dwrap,relpath)
      print("== Creating wrappers in:",outwrapdir)

      if outwrapdir != "":
        if not dry:
          pathlib.Path(outwrapdir).mkdir(parents=True, exist_ok=True)
        else:
          print("mkdir -p",outwrapdir)

      for name in os.listdir(path):
        f = os.path.join(path,name)
        if os.access(f,os.X_OK):
          outwrap = os.path.join(outwrapdir,name)
          if not os.path.islink(outwrap):
            print("Creating wrapper:",outwrap,"for",f)
            if not dry:
              os.symlink(wrapper,outwrap)
            else:
              print("ln -s",wrapper,outwrap)
          else:
            print("Wrapper already exists:",outwrap)

def process_module_dir(md, d, dout, dwrap, wrapper, tag, dry=False):
    modulename = os.path.basename(d)
    for name in os.listdir(d):
        if name.endswith(".lua") and name != "recommendation.lua":
          process_module_dir(md, d, dout, dwrap, wrapper, name, tag, dry=dry)

if __name__ == "__main__":

    hpc_sysname = os.getenv("HPC_SSE_SYSNAME")
    if hpc_sysname is None or hpc_sysname == "":
      print("Missing HPC_SSE_SYSNAME environment variable")
      exit(1)

    parser = argparse.ArgumentParser(description="Create a SSE2 compat wrap module", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('module', help='The SSE1 module to wrap.')
    parser.add_argument('--indir', help='The path in which to look for modules to wrap.', default="/software/sse/modules")
    parser.add_argument('--outdir', help='The path to create the wrapped compatibiltiy modules in.', default="/software/sse2/compat/modules")
    parser.add_argument('--wrapdir', help='The path for the wrapper links.', default="/software/sse2/compat/wrap/")
    parser.add_argument('--wrapper', help='The wrapper script that the links will link to.', default="/software/sse2/generic/manual/ssetools/latest/compat/hpc_compat_el7_auto")
    parser.add_argument('--tag', help='Tag the module with this flag to clarify that it is a compatibility module.', default="-compat-el7")
    parser.add_argument('--dry', help='Dry run', action='store_true')
    parser.set_defaults(dry=False)
    args = parser.parse_args()

    module_path = os.path.join(args.indir, args.module)
    
    if os.path.exists(module_path+".lua"):
      module_dir = os.path.dirname(module_path)
      module_name = os.path.basename(module_path+".lua")
      process_module(args.indir, module_dir, args.outdir, args.wrapdir, args.wrapper, args.tag, module_name, dry=args.dry)
    elif os.path.exists(module_path):
      process_module_dir(args.indir, module_path, args.outdir, args.wrapdir, args.wrapper, args.tag, dry=args.dry)
    else:
      print("The specified module: "+module_path+" does not exist.")
      exit(1)
      
