#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version

set -e

TEST_SCRIPT_DIR_REL="$(dirname -- "${BASH_SOURCE[0]}")"
TEST_SCRIPT_DIR="$(cd -- "${TEST_SCRIPT_DIR_REL}" && pwd)"

echo "Note: you are meant to load an hpc module before invoking this test script and give the ABINIT software module as an argument"
echo

ABINITMODULE="$1"

if [ -z "$ABINITMODULE" ]; then
    echo "Must specify an Abinit software module"
    exit 1
fi

export TESTDIR=$(mktemp -d --tmpdir=.)
cd "$TESTDIR"

cp -rp "${TEST_SCRIPT_DIR_REL}/abinit_test" .
cd abinit_test

echo
echo "== Executing ABINIT interactively on one node"

OUTPUT=$(interactive -A nsc --reserv now -t 00:02:00 -N 1 <<< "module load $ABINITMODULE"$'\nmpprun abinit mpiio_t01.abi > abinit.out 2>&1\nexit' | grep "\S" | sort)
echo "---------"
echo "$OUTPUT"
echo "---------"

SHA256=$(echo "$OUTPUT"|sha256sum)

if [ "$SHA256" != "01ba4719c80b6fe911b091a7c05124b64eeece964e09c058ef8f9805daca546b  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

ENERGY=$(cat abinit.out | grep ETOT | tail -n 1 | awk '{print $3}')
ROUNDENERGY=$(LC_ALL=C /usr/bin/printf '%.*f\n' 11 "$ENERGY")
if [ "$ROUNDENERGY" != "-8.59387312151" ]; then
    echo "ERROR: Calculation did not give the right answer."
    exit 1
fi
rm -f abinit.out

echo
echo "== Executing ABINIT interactively on two nodes"

OUTPUT=$(interactive -A nsc --reserv now -t 00:02:00 -N 2 <<< "module load $ABINITMODULE"$'\nmpprun abinit mpiio_t01.abi > abinit.out 2>&1\nexit' | grep "\S" | sort)
echo "---------"
echo "$OUTPUT"
echo "---------"

SHA256=$(echo "$OUTPUT"|sha256sum)

if [ "$SHA256" != "01ba4719c80b6fe911b091a7c05124b64eeece964e09c058ef8f9805daca546b  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

ENERGY=$(cat abinit.out | grep ETOT | tail -n 1 | awk '{print $3}')
ROUNDENERGY=$(LC_ALL=C /usr/bin/printf '%.*f\n' 11 "$ENERGY")
if [ "$ROUNDENERGY" != "-8.59387312151" ]; then
    echo "ERROR: Calculation did not give the right answer."
    exit 1
fi
rm -f abinit.out

echo
echo "== Executing ABINIT in batch mode on one node"
sbatch --wait -o slurm.out -e slurm.err <<EOF
#!/bin/bash
#SBATCH -A nsc
#SBATCH --reserv now
#SBATCH -N 1
#SBATCH -t 0:02:00

module list >&2

module load $ABINITMODULE
mpprun abinit mpiio_t01.abi > abinit.out
EOF
cat slurm.err
cat slurm.out

SHA256=$(cat slurm.out | grep "\S" | sort | sha256sum)

if [ "$SHA256" != "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

ENERGY=$(cat abinit.out | grep ETOT | tail -n 1 | awk '{print $3}')
ROUNDENERGY=$(LC_ALL=C /usr/bin/printf '%.*f\n' 11 "$ENERGY")
if [ "$ROUNDENERGY" != "-8.59387312151" ]; then
    echo "ERROR: Calculation did not give the right answer."
    exit 1
fi
rm -f abinit.out

echo
echo "== Executing ABINIT in batch mode on two nodes"
sbatch --wait -o slurm.out -e slurm.err <<EOF
#!/bin/bash
#SBATCH -A nsc
#SBATCH --reserv now
#SBATCH -N 2
#SBATCH -t 0:02:00

module list >&2

module load $ABINITMODULE
mpprun abinit mpiio_t01.abi > abinit.out
EOF
cat slurm.err
cat slurm.out

SHA256=$(cat slurm.out | grep "\S" | sort | sha256sum)

if [ "$SHA256" != "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

ENERGY=$(cat abinit.out | grep ETOT | tail -n 1 | awk '{print $3}')
ROUNDENERGY=$(LC_ALL=C /usr/bin/printf '%.*f\n' 11 "$ENERGY")
if [ "$ROUNDENERGY" != "-8.59387312151" ]; then
    echo "ERROR: Calculation did not give the right answer."
    exit 1
fi
rm -f abinit.out

echo
echo "== All ABINIT software tests passed =="
exit 0
