#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version

set -e

TEST_SCRIPT_DIR_REL="$(dirname -- "${BASH_SOURCE[0]}")"
TEST_SCRIPT_DIR="$(cd -- "${TEST_SCRIPT_DIR_REL}" && pwd)"

echo "Note: you are meant to load an hpc module before invoking this test script and give the container, buildenv module, and compiler binary name inside container to test as an argument"
echo
echo "== Building a simple MPI test program inside the el7 compatibilty container"

CONTAINER=$1
BUILDENVMODULE="$2"
COMPILER="$3"

if [ -z "$CONTAINER" -o -z "$BUILDENVMODULE" -o -z "$COMPILER" ]; then
    echo "Must specify container name, buildenv module, and compiler to this test script"
    exit 1
fi

export TESTDIR=$(mktemp -d --tmpdir=.)
cd "$TESTDIR"

cp -rp "${TEST_SCRIPT_DIR_REL}/simple_mpi_test" .
(
    cd simple_mpi_test;
    MPICC="hpc_container_exec $CONTAINER -m $BUILDENVMODULE $COMPILER" make
)

echo
echo "== Executing simple MPI test program interactively on one node"

OUTPUT=$(interactive -A nsc --reserv now -t 00:02:00 -N 1 <<< "mpprun --compat $CONTAINER simple_mpi_test/simple_mpi_test"$'\n'"exit" | grep "\S" | sort)
echo "---------"
echo "$OUTPUT"
echo "---------"

SHA256=$(echo "$OUTPUT"|sha256sum)

if [ "$SHA256" != "3f4b14b2d77b5beaef354612efc40b53b39c8b9ef77c065f9a547eac6e1021e1  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

echo
echo "== Executing simple MPI test program on two nodes"

OUTPUT=$(interactive -A nsc --reserv now -t 00:02:00 -N 2 <<< "mpprun --compat $CONTAINER simple_mpi_test/simple_mpi_test"$'\n'"exit" | grep "\S" | sort)
echo "$OUTPUT"

SHA256=$(echo "$OUTPUT"|sha256sum)

if [ "$SHA256" != "309e7d2bd95a7e88e6b9f07135a0863aa5746adce0de2657a2b00cd9a455d6fe  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

echo
echo "== Executing simple MPI test program in batch mode on one node"
sbatch --wait -o slurm.out -e slurm.err <<EOF
#!/bin/bash
#SBATCH -A nsc
#SBATCH --reserv now
#SBATCH -N 1
#SBATCH -t 0:02:00

module list >&2

mpprun --compat $CONTAINER simple_mpi_test/simple_mpi_test
EOF
cat slurm.err
cat slurm.out

SHA256=$(cat slurm.out | grep "\S" | sort | sha256sum)

if [ "$SHA256" != "3f4b14b2d77b5beaef354612efc40b53b39c8b9ef77c065f9a547eac6e1021e1  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

echo
echo "== Executing simple MPI test program in batch mode on two nodes"
sbatch --wait -o slurm.out -e slurm.err <<EOF
#!/bin/bash
#SBATCH -A nsc
#SBATCH --reserv now
#SBATCH -N 2
#SBATCH -t 0:02:00

module list >&2

mpprun --compat $CONTAINER simple_mpi_test/simple_mpi_test
EOF
cat slurm.err
cat slurm.out

SHA256=$(cat slurm.out | grep "\S" | sort | sha256sum)

if [ "$SHA256" != "309e7d2bd95a7e88e6b9f07135a0863aa5746adce0de2657a2b00cd9a455d6fe  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

echo
echo "== All compat tests passed =="
exit 0
