#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version

set -e

TEST_SCRIPT_DIR_REL="$(dirname -- "${BASH_SOURCE[0]}")"
TEST_SCRIPT_DIR="$(cd -- "${TEST_SCRIPT_DIR_REL}" && pwd)"

echo "(Note: you are meant to load an hpc module before invoking this test script and give the Elk software module as an argument)"
echo

ELKMODULE="$1"

if [ -z "$ELKMODULE" ]; then
    echo "Must specify an Elk software module"
    exit 1
fi

export TESTDIR=$(mktemp -d --tmpdir=.)
cd "$TESTDIR"

cp -rp "${TEST_SCRIPT_DIR_REL}/elk_test" .
cd elk_test

echo
echo "== Executing Elk interactively on one node"

OUTPUT=$(interactive -A nsc --reserv now -t 00:02:00 -N 1 <<< "module load $ELKMODULE"$'\nmpprun elk\nexit' | grep "\S" | sort)
echo "---------"
echo "$OUTPUT"
echo "---------"

SHA256=$(echo "$OUTPUT"|sha256sum)

if [ "$SHA256" != "b11a5a99817b517d0ccfecc17983b29c3aec476bbd8c3a1bfcc8afcf0f111013  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

ENERGY="$(cat TOTENERGY.OUT | tail -n 1 | sed -e 's/  *$//')"
ROUNDENERGY=$(LC_ALL=C /usr/bin/printf '%.*f\n' 8 "$ENERGY")
if [ "$ROUNDENERGY" != "-967.67400991" ]; then
    echo "ERROR: Calculation did not give the right answer."
    exit 1
fi

echo
echo "== Executing Elk interactively on two nodes"

OUTPUT=$(interactive -A nsc --reserv now -t 00:02:00 -N 2 <<< "module load $ELKMODULE"$'\nmpprun elk\nexit' | grep "\S" | sort)
echo "---------"
echo "$OUTPUT"
echo "---------"

SHA256=$(echo "$OUTPUT"|sha256sum)

if [ "$SHA256" != "d6e96d88e35b7f836dcbd05f0767ccc117a59c5f2396e608551f62288274caa8  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

ENERGY="$(cat TOTENERGY.OUT | tail -n 1 | sed -e 's/  *$//')"
ROUNDENERGY=$(LC_ALL=C /usr/bin/printf '%.*f\n' 8 "$ENERGY")
if [ "$ROUNDENERGY" != "-967.67400991" ]; then
    echo "ERROR: Calculation did not give the right answer."
    exit 1
fi

echo
echo "== Executing Elk in batch mode on one node"
sbatch --wait -o slurm.out -e slurm.err <<EOF
#!/bin/bash
#SBATCH -A nsc
#SBATCH --reserv now
#SBATCH -N 1
#SBATCH -t 0:02:00

module list >&2

module load $ELKMODULE
mpprun elk
EOF
cat slurm.err
cat slurm.out

SHA256=$(cat slurm.out | grep "\S" | sort | sha256sum)

if [ "$SHA256" != "b11a5a99817b517d0ccfecc17983b29c3aec476bbd8c3a1bfcc8afcf0f111013  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

ENERGY="$(cat TOTENERGY.OUT | tail -n 1 | sed -e 's/  *$//')"
ROUNDENERGY=$(LC_ALL=C /usr/bin/printf '%.*f\n' 8 "$ENERGY")
if [ "$ROUNDENERGY" != "-967.67400991" ]; then
    echo "ERROR: Calculation did not give the right answer."
    exit 1
fi

echo
echo "== Executing Elk in batch mode on two nodes"
sbatch --wait -o slurm.out -e slurm.err <<EOF
#!/bin/bash
#SBATCH -A nsc
#SBATCH --reserv now
#SBATCH -N 2
#SBATCH -t 0:02:00

module list >&2

module load $ELKMODULE
mpprun elk
EOF
cat slurm.err
cat slurm.out

SHA256=$(cat slurm.out | grep "\S" | sort | sha256sum)

if [ "$SHA256" != "d6e96d88e35b7f836dcbd05f0767ccc117a59c5f2396e608551f62288274caa8  -" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

ENERGY="$(cat TOTENERGY.OUT | tail -n 1 | sed -e 's/  *$//')"
ROUNDENERGY=$(LC_ALL=C /usr/bin/printf '%.*f\n' 8 "$ENERGY")
if [ "$ROUNDENERGY" != "-967.67400991" ]; then
    echo "ERROR: Calculation did not give the right answer."
    exit 1
fi

echo
echo "== All Elk software tests passed =="
exit 0
