#!/bin/bash
#
# This file is part of ssetools
# Copyright © 2023-2024 The ssetools AUTHORS
#
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file of the ssetools repository.
#
# History:
#   rar@nsc: 2023 initial version

set -e

desired_output() {
  broadcast_message=99
  sent_message=42
  ranks=$1
  echo "Broadcasting message: $broadcast_message"
  echo "Message Sent: $sent_message"
  echo "Process rank 0 received broadcasted message: $broadcast_message"
  for rank in $(seq 1 "$((ranks-1))"); do
    echo "Process rank $rank received broadcasted message: $broadcast_message"
    echo "Process rank $rank received: $sent_message"
  done
}

TEST_SCRIPT_DIR_REL="$(dirname -- "${BASH_SOURCE[0]}")"
TEST_SCRIPT_DIR="$(cd -- "${TEST_SCRIPT_DIR_REL}" && pwd)"

# Compose slurm arguments and sbatch headers from TEST_SLURM_ALLOC and TEST_SLURM_RESERV if set
TEST_SLURM_ARGS=""
TEST_SBATCH_HEADER="#!/bin/bash"
if [ -n "$TEST_SLURM_ALLOC" ]; then
    if [ -n "$TEST_SLURM_RESERV" ]; then
	TEST_SLURM_ARGS="-A $TEST_SLURM_ALLOC --reserv $TEST_SLURM_RESERV"
	TEST_SBATCH_HEADER="#!/bin/bash
#SBATCH -A $TEST_SLURM_ALLOC
#SBATCH --reserv $TEST_SLURM_RESERV"
    else
	TEST_SLURM_ARGS="-A $TEST_SLURM_ALLOC --reserv $TEST_SLURM_RESERV"
	TEST_SBATCH_HEADER="#!/bin/bash
#SBATCH -A $TEST_SLURM_ALLOC"
    fi
fi

echo "Note: you are meant to load an hpc module before invoking this test script and give the buildenv module, and compiler binary name inside container to test as an argument"
echo
echo "== Building a simple MPI test program using the loaded toolchain"

BUILDENVMODULE="$1"
COMPILER="$2"
CORES_PER_NODE="$3"
if [ -n "$3" ]; then
    CORES_PER_NODE="$3"
else
    CORES_PER_NODE="32"
fi

if [ -z "$BUILDENVMODULE" -o -z "$COMPILER" ]; then
    echo "Must specify buildenv module, and compiler to this test script"
    exit 1
fi

export TESTDIR=$(mktemp -d --tmpdir=.)
cd "$TESTDIR"

cp -rp "${TEST_SCRIPT_DIR_REL}/simple_mpi_test" .

(
    module load "$BUILDENVMODULE"
    cd simple_mpi_test
    MPICC="$COMPILER" make
)

echo
echo "== Executing simple MPI test program interactively on one node"

OUTPUT=$(interactive $TEST_SLURM_ARGS -t 00:02:00 -N 1 <<< $'mpprun simple_mpi_test/simple_mpi_test\nexit' | grep "\S" | sort)
echo "---------"
echo "$OUTPUT"
echo "---------"

SHA256=$(echo "$OUTPUT"|sha256sum)
SHA256_CORRECT=$(desired_output "$CORES_PER_NODE" | sort | sha256sum)

if [ "$SHA256" != "$SHA256_CORRECT" ]; then
    echo "ERROR: Output does not look correct"
    echo "  $SHA256"
    echo "  vs."
    echo "  $SHA256_CORRECT"
    exit 1
fi

echo
echo "== Executing simple MPI test program on two nodes"

OUTPUT=$(interactive $TEST_SLURM_ARGS -t 00:02:00 -N 2 <<< $'mpprun simple_mpi_test/simple_mpi_test\nexit' | grep "\S" | sort)
echo "---------"
echo "$OUTPUT"
echo "---------"

SHA256=$(echo "$OUTPUT"|sha256sum)
SHA256_CORRECT=$(desired_output "$((2*CORES_PER_NODE))" | sort | sha256sum)

if [ "$SHA256" != "$SHA256_CORRECT" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

echo
echo "== Executing simple MPI test program in batch mode on one node"

cat - <<EOF
$TEST_SBATCH_HEADER
#SBATCH -N 2
#SBATCH -t 0:02:00

module list >&2

mpprun simple_mpi_test/simple_mpi_test
EOF

sbatch --wait -o slurm.out -e slurm.err <<EOF
$TEST_SBATCH_HEADER
#SBATCH -N 1
#SBATCH -t 0:02:00

module list >&2

mpprun simple_mpi_test/simple_mpi_test
EOF
cat slurm.err
cat slurm.out

SHA256=$(cat slurm.out | grep "\S" | sort | sha256sum)
SHA256_CORRECT=$(desired_output "$CORES_PER_NODE" | sort | sha256sum)

if [ "$SHA256" != "$SHA256_CORRECT" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

echo
echo "== Executing simple MPI test program in batch mode on two nodes"

sbatch --wait -o slurm.out -e slurm.err <<EOF
$TEST_SBATCH_HEADER
#SBATCH -N 2
#SBATCH -t 0:02:00

module list >&2

mpprun simple_mpi_test/simple_mpi_test
EOF
cat slurm.err
cat slurm.out

SHA256=$(cat slurm.out | grep "\S" | sort | sha256sum)
SHA256_CORRECT=$(desired_output "$((2*CORES_PER_NODE))" | sort | sha256sum)

if [ "$SHA256" != "$SHA256_CORRECT" ]; then
    echo "ERROR: Output does not look correct ($SHA256)"
    exit 1
fi

echo
echo "== All toolchain tests passed =="
exit 0
